@@ -2,32 +2,7 @@
  * Copyright (c) 2002, 2003, 2004 Niels Provos <provos@citi.umich.edu>
  * All rights reserved.
  * 
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * <LICENSEHERE>
  */
 
 #include <sys/types.h>
@@ -47,6 +22,7 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <signal.h>
+#define _GNU_SOURCE
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -69,6 +45,8 @@
 #include "fdpass.h"
 #include "osfp.h"
 #include "pyextend.h"
+#include "honeyd_overload.h"
+#include "util.h"
 
 ssize_t atomicio(ssize_t (*)(), int, void *, size_t);
 
@@ -81,7 +59,7 @@ cmd_trigger_read(struct command *cmd, int size)
  	if (cmd->pfd == -1 || !cmd->fdconnected)
 		return;
 	if (size)
-		event_add(&cmd->pread, NULL);
+		TRACE(cmd->pread.ev_fd, event_add(&cmd->pread, NULL));
 }
 
 void
@@ -90,20 +68,21 @@ cmd_trigger_write(struct command *cmd, int size)
  	if (cmd->pfd == -1 || !cmd->fdconnected)
 		return;
 	if (size)
-		event_add(&cmd->pwrite, NULL);
+		TRACE(cmd->pwrite.ev_fd, event_add(&cmd->pwrite, NULL));
 }
 
 void
 cmd_free(struct command *cmd)
 {
-	event_del(&cmd->pread);
-	event_del(&cmd->pwrite);
-	close(cmd->pfd);
+	TRACE(cmd->pread.ev_fd, event_del(&cmd->pread));
+	TRACE(cmd->pwrite.ev_fd, event_del(&cmd->pwrite));
+	TRACE_RESET(cmd->pfd, close(cmd->pfd));
 	cmd->pfd = -1;
+	cmd->pid = -1;
 
 	if (cmd->perrfd != -1) {
-		event_del(&cmd->peread);
-		close(cmd->perrfd);
+		TRACE(cmd->peread.ev_fd, event_del(&cmd->peread));
+		TRACE_RESET(cmd->perrfd, close(cmd->perrfd));
 		cmd->perrfd = -1;
 	}
 
@@ -116,13 +95,17 @@ cmd_free(struct command *cmd)
 void
 cmd_ready_fd(struct command *cmd, struct callback *cb, void *con)
 {
-	event_set(&cmd->pread, cmd->pfd, EV_READ, cb->cb_read, con);
-	event_set(&cmd->pwrite, cmd->pfd, EV_WRITE, cb->cb_write, con);
+	TRACE(cmd->pfd,
+	    event_set(&cmd->pread, cmd->pfd, EV_READ, cb->cb_read, con));
+	TRACE(cmd->pfd,
+	    event_set(&cmd->pwrite, cmd->pfd, EV_WRITE, cb->cb_write, con));
 	cmd->fdconnected = 1;
 
-	if (cmd->perrfd != -1)
-		event_set(&cmd->peread, cmd->perrfd, EV_READ, cb->cb_eread,
-		    con);
+	if (cmd->perrfd != -1) {
+		TRACE(cmd->perrfd,
+		    event_set(&cmd->peread, cmd->perrfd, EV_READ, cb->cb_eread,
+			con));
+	}
 }
 
 struct addrinfo *
@@ -171,7 +154,7 @@ cmd_proxy_connect(struct tuple *hdr, struct command *cmd, struct addrinfo *ai,
         if (fcntl(fd, F_SETFD, 1) == -1)
                 warn("fcntl(F_SETFD)");
 
-	cmd->pfd = fd;
+	TRACE(fd, cmd->pfd = fd);
         if (connect(fd, ai->ai_addr, ai->ai_addrlen) == 0) {
 		(*cb->cb_connect)(fd, EV_WRITE, con);
 		return (0);
@@ -180,12 +163,12 @@ cmd_proxy_connect(struct tuple *hdr, struct command *cmd, struct addrinfo *ai,
 	if (errno != EINPROGRESS) {
 		warn("connect");
 		cmd->pfd = -1;
-		close(fd);
+		TRACE_RESET(fd, close(fd));
 		return (-1);
 	}
 
-	event_set(&cmd->pwrite, fd, EV_WRITE, cb->cb_connect, con);
-	event_add(&cmd->pwrite, &tv);
+	TRACE(fd, event_set(&cmd->pwrite, fd, EV_WRITE, cb->cb_connect, con));
+	TRACE(cmd->pwrite.ev_fd, event_add(&cmd->pwrite, &tv));
 
 	if (getnameinfo(ai->ai_addr, ai->ai_addrlen,
 		ntop, sizeof(ntop), strport, sizeof(strport),
@@ -208,13 +191,6 @@ cmd_environment(struct template *tmpl, struct tuple *hdr)
 	struct ip_hdr ip;
 	char *os_name;
 
-	/* Determine the remote operating system */
-	ip.ip_src = hdr->ip_src;
-	os_name = honeyd_osfp_name(&ip);
-	if (os_name != NULL) {
-		setenv("HONEYD_REMOTE_OS", os_name, 1);
-	}
-
 	if (tmpl->person != NULL) {
 		snprintf(line, sizeof(line), "%s", tmpl->person->name);
 		setenv("HONEYD_PERSONALITY", line, 1);
@@ -223,6 +199,13 @@ cmd_environment(struct template *tmpl, struct tuple *hdr)
 	if (hdr == NULL)
 		return;
 	     
+	/* Determine the remote operating system */
+	ip.ip_src = hdr->ip_src;
+	os_name = honeyd_osfp_name(&ip);
+	if (os_name != NULL) {
+		setenv("HONEYD_REMOTE_OS", os_name, 1);
+	}
+
 	addr_pack(&addr, ADDR_TYPE_IP, IP_ADDR_BITS, &hdr->ip_src,IP_ADDR_LEN);
 	snprintf(line, sizeof(line), "%s", addr_ntoa(&addr));
 	setenv("HONEYD_IP_SRC", line, 1);
@@ -358,8 +341,8 @@ cmd_fork(struct tuple *hdr, struct command *cmd, struct template *tmpl,
 	if (socketpair(AF_UNIX, hdr->type, 0, pair) == -1)
 		return (-1);
 	if (socketpair(AF_UNIX, SOCK_STREAM, 0, perr) == -1) {
-		close(pair[0]);
-		close(pair[1]);
+		TRACE_RESET(pair[0], close(pair[0]));
+		TRACE_RESET(pair[1], close(pair[1]));
 		return (-1);
 	}
 
@@ -382,17 +365,17 @@ cmd_fork(struct tuple *hdr, struct command *cmd, struct template *tmpl,
 		cmd_setpriv(tmpl);
 
 		/* Child */
-		close(pair[0]);
+		TRACE_RESET(pair[0], close(pair[0]));
 		if (dup2(pair[1], fileno(stdout)) == -1)
 			err(1, "%s: dup2", __func__);
 		if (dup2(pair[1], fileno(stdin)) == -1)
 			err(1, "%s: dup2", __func__);
-		close(perr[0]);
+		TRACE_RESET(pair[0], close(perr[0]));
 		if (dup2(perr[1], fileno(stderr)) == -1)
 			err(1, "%s: dup2", __func__);
 
-		close(pair[1]);
-		close(perr[1]);
+		TRACE_RESET(pair[1], close(pair[1]));
+		TRACE_RESET(perr[1], close(perr[1]));
 
 		cmd_environment(tmpl, hdr);
 
@@ -402,14 +385,14 @@ cmd_fork(struct tuple *hdr, struct command *cmd, struct template *tmpl,
 		/* NOT REACHED */
 	}
 
-	close(pair[1]);
-	cmd->pfd = pair[0];
+	TRACE_RESET(pair[1], close(pair[1]));
+	TRACE(pair[0], cmd->pfd = pair[0]);
 	if (fcntl(cmd->pfd, F_SETFD, 1) == -1)
 		warn("fcntl(F_SETFD)");
 	if (fcntl(cmd->pfd, F_SETFL, O_NONBLOCK) == -1)
 		warn("fcntl(F_SETFL)");
 
-	close(perr[1]);
+	TRACE_RESET(perr[1], close(perr[1]));
 	cmd->perrfd = perr[0];
 	if (fcntl(cmd->perrfd, F_SETFD, 1) == -1)
 		warn("fcntl(F_SETFD)");
@@ -423,8 +406,8 @@ cmd_fork(struct tuple *hdr, struct command *cmd, struct template *tmpl,
 
 	cmd_ready_fd(cmd, cb, con);
 
-	event_add(&cmd->pread, NULL);
-	event_add(&cmd->peread, NULL);
+	TRACE(cmd->pread.ev_fd, event_add(&cmd->pread, NULL));
+	TRACE(cmd->peread.ev_fd, event_add(&cmd->peread, NULL));
 
 	honeyd_nchildren++;
 
@@ -442,10 +425,10 @@ cmd_fork(struct tuple *hdr, struct command *cmd, struct template *tmpl,
 		warn("sigprocmask");
 
  fork_err:
-	close(perr[0]);
-	close(perr[1]);
-	close(pair[0]);
-	close(pair[1]);
+	TRACE_RESET(perr[0], close(perr[0]));
+	TRACE_RESET(perr[1], close(perr[1]));
+	TRACE_RESET(pair[0], close(pair[0]));
+	TRACE_RESET(pair[1], close(pair[1]));
 	cmd->pfd = -1;
 
 	return (-1);
@@ -460,7 +443,7 @@ cmd_python(struct tuple *hdr, struct command *cmd, void *con)
 	if (socketpair(AF_UNIX, hdr->type, 0, pair) == -1)
 		return (-1);
 
-	cmd->pfd = pair[0];
+	TRACE(pair[0], cmd->pfd = pair[0]);
 	if (fcntl(cmd->pfd, F_SETFD, 1) == -1)
 		warn("fcntl(F_SETFD)");
 	if (fcntl(cmd->pfd, F_SETFL, O_NONBLOCK) == -1)
@@ -479,7 +462,7 @@ cmd_python(struct tuple *hdr, struct command *cmd, void *con)
 
 	cmd_ready_fd(cmd, cb, con);
 
-	event_add(&cmd->pread, NULL);
+	TRACE(cmd->pread.ev_fd, event_add(&cmd->pread, NULL));
 
 	return (pair[1]);
 }
@@ -512,19 +495,24 @@ cmd_subsystem(struct template *tmpl, struct subsystem *sub,
 	}
 
 	if (cmd->pid == 0) {
+		char magic_buf[12];
+		int magic_fd;
 		/* Set privileges */
 		cmd_setpriv(tmpl);
 
 		/* Child */
-		close(pair[0]);
-		if (dup2(pair[1], SUBSYSTEM_MAGICFD) == -1)
-			err(1, "%s: dup2", __func__);
+		TRACE_RESET(pair[0], close(pair[0]));
+		/* Set the communication fd */
+		if ((magic_fd = dup(pair[1])) == -1)
+			err(1, "%s: dup(%d): no magic", __func__, pair[1]);
+		snprintf(magic_buf, sizeof(magic_buf), "%d", magic_fd);
+		setenv(SUBSYSTEM_MAGICFD, magic_buf, 1);
 		if (dup2(fileno(stderr), fileno(stdout)) == -1)
 			err(1, "%s: dup2", __func__);
 		if (dup2(fileno(stderr), fileno(stdin)) == -1)
 			err(1, "%s: dup2", __func__);
 
-		close(pair[1]);
+		TRACE_RESET(pair[1], close(pair[1]));
 
 		cmd_environment(tmpl, NULL);
 
@@ -538,17 +526,15 @@ cmd_subsystem(struct template *tmpl, struct subsystem *sub,
 		/* NOT REACHED */
 	}
 
-	close(pair[1]);
-	cmd->pfd = pair[0];
+	TRACE_RESET(pair[1], close(pair[1]));
+	TRACE(pair[0], cmd->pfd = pair[0]);
 	if (fcntl(cmd->pfd, F_SETFD, 1) == -1)
 		warn("fcntl(F_SETFD)");
-	if (fcntl(cmd->pfd, F_SETFL, O_NONBLOCK) == -1)
-		warn("fcntl(F_SETFL)");
 
 	cmd->perrfd = -1;
 	cmd_ready_fd(cmd, &subsystem_cb, sub);
 
-	event_add(&cmd->pread, NULL);
+	TRACE(cmd->pread.ev_fd, event_add(&cmd->pread, NULL));
 
 	honeyd_nchildren++;
 
@@ -566,21 +552,73 @@ cmd_subsystem(struct template *tmpl, struct subsystem *sub,
 		warn("sigprocmask");
 
  fork_err:
-	close(pair[0]);
-	close(pair[1]);
+	TRACE_RESET(pair[0], close(pair[0]));
+	TRACE_RESET(pair[1], close(pair[1]));
 	cmd->pfd = -1;
 
 	return (-1);
 }
 
+static void
+cmd_subsystem_connect_cb(int fd, short what, void *arg)
+{
+	struct port_encapsulate *tmp = arg;
+	struct port *port = tmp->port;
+
+	TAILQ_REMOVE(&port->pending, tmp, next);
+
+	if (what != EV_WRITE) {
+		/* We encountered some error with this */
+		if (tmp->hdr->type == SOCK_STREAM)
+			tcp_connectfail(tmp->con);
+		goto out;
+	}
+
+	cmd_subsystem_connect(tmp->hdr, tmp->cmd, port, tmp->con);
+
+ out:
+	port_encapsulation_free(tmp);
+}
+
+int
+cmd_subsystem_schedule_connect(struct tuple *hdr, struct command *cmd,
+    struct port *port, void *con)
+{
+	struct port_encapsulate *tmp = calloc(1, sizeof(*tmp));
+	struct subsystem *sub = port->sub;
+
+	if (tmp == NULL)
+		return (-1);
+
+	tmp->hdr = hdr;
+	tmp->cmd = cmd;
+	tmp->port = port;
+	tmp->con = con;
+
+	/* Tell the connection that it has a pending connection */
+	tmp->hdr->pending = tmp;
+
+	TAILQ_INSERT_TAIL(&port->pending, tmp, next);
+
+	TRACE(port->sub_fd, event_set(&tmp->ev, port->sub_fd, EV_WRITE,
+	    cmd_subsystem_connect_cb, tmp));
+	TRACE(tmp->ev.ev_fd, event_add(&tmp->ev, NULL));
+
+	syslog(LOG_DEBUG,
+	    "Scheduling connection establishment: %s -> subsystem \"%s\"",
+	    honeyd_contoa(hdr), sub->cmdstring);
+
+	return (0);
+}
+
 int
 cmd_subsystem_connect(struct tuple *hdr, struct command *cmd,
     struct port *port, void *con)
 {
 	struct callback *cb;
 	struct subsystem *sub = port->sub;
-	struct sockaddr_in si;
-	struct addr src;
+	struct bundle bundle;
+	struct addr src, dst;
 	int pair[2];
         
 	if (hdr->type == SOCK_STREAM)
@@ -589,7 +627,7 @@ cmd_subsystem_connect(struct tuple *hdr, struct command *cmd,
 		cb = &cb_udp;
 
         if (socketpair(AF_LOCAL, hdr->type, 0, pair) == -1) {
-                warn("socketpair");
+                warn("%s: socketpair: %s", __func__, sub->cmdstring);
                 return (-1);
         }
 
@@ -599,18 +637,25 @@ cmd_subsystem_connect(struct tuple *hdr, struct command *cmd,
         if (fcntl(pair[0], F_SETFD, 1) == -1)
                 warn("fcntl(F_SETFD)");
 
-	cmd->pfd = pair[0];
+	TRACE(pair[0], cmd->pfd = pair[0]);
 
-	/* Prepare sockaddr */
+	/* Prepare sockaddr for both src and destination */
 	addr_pack(&src, ADDR_TYPE_IP, IP_ADDR_BITS, &hdr->ip_src,IP_ADDR_LEN);
-	addr_ntos(&src, (struct sockaddr *)&si);
-	si.sin_port = htons(hdr->sport);
+	addr_ntos(&src, (struct sockaddr *)&bundle.src);
+	bundle.src.sin_port = htons(hdr->sport);
+	addr_pack(&dst, ADDR_TYPE_IP, IP_ADDR_BITS, &hdr->ip_dst,IP_ADDR_LEN);
+	addr_ntos(&dst, (struct sockaddr *)&bundle.dst);
+	bundle.dst.sin_port = htons(hdr->dport);
 
-	if (send_fd(port->sub_fd, pair[1], &si, sizeof(si)) == -1)
+	if (send_fd(port->sub_fd, pair[1], &bundle, sizeof(bundle)) == -1) {
+		TRACE_RESET(pair[0], close(pair[0]));
+		TRACE_RESET(pair[1], close(pair[1]));
+		cmd->pfd = -1;
 		return (-1);
+	}
 
 	/* After transfering the file descriptor, we may close it */
-	close(pair[1]);
+	TRACE_RESET(pair[1], close(pair[1]));
 
 	/* We are connected now */
 	(*cb->cb_connect)(pair[0], EV_WRITE, con);
@@ -657,16 +702,18 @@ cmd_subsystem_localconnect(struct tuple *hdr, struct command *cmd,
 
 		/* Confirm success of failure */
 		res = fd == -1 ? -1 : 0;
-		atomicio(write, sub->cmd.pfd, &res, 1);
+		TRACE(sub->cmd.pfd,
+		    atomicio(write, sub->cmd.pfd, &res, 1));
 		if (fd == -1)
 			return (-1);
 
-		port->sub_fd = fd;
+		TRACE(fd, port->sub_fd = fdshare_dup(fd));
 	}
 
 	/* Get another fd on this special thingy */
 	while ((fd = receive_fd(port->sub_fd, NULL, NULL)) == -1) {
 		if (errno != EAGAIN) {
+			TRACE(port->sub_fd, fdshare_close(port->sub_fd));
 			warnx("%s: receive_fd", __func__);
 			return (-1);
 		}
@@ -678,7 +725,7 @@ cmd_subsystem_localconnect(struct tuple *hdr, struct command *cmd,
         if (fcntl(fd, F_SETFD, 1) == -1)
                 warn("%s: fcntl(F_SETFD)", __func__);
 
-	cmd->pfd = fd;
+	TRACE(fd, cmd->pfd = fd);
 
 	/* Prepare sockaddr */
 	addr_pack(&src, ADDR_TYPE_IP, IP_ADDR_BITS, &hdr->ip_dst, IP_ADDR_LEN);
@@ -686,11 +733,16 @@ cmd_subsystem_localconnect(struct tuple *hdr, struct command *cmd,
 	si.sin_port = htons(hdr->dport);
 
 	/* Write the bound socket address to the other side */
-	if (atomicio(write, port->sub_fd, &si, sizeof(si)) != sizeof(si))
+	if (atomicio(write, port->sub_fd, &si, sizeof(si)) != sizeof(si)) {
+		TRACE(port->sub_fd, fdshare_close(port->sub_fd));
+		port->sub_fd = -1;
+		TRACE_RESET(cmd->pfd, close(cmd->pfd));
+		cmd->pfd = -1;
 		return (-1);
+	}
 
 	/* Now we may close the special thingy */
-	close(port->sub_fd);
+	TRACE(port->sub_fd, fdshare_close(port->sub_fd));
 	port->sub_fd = -1;
 
 	/* We are connected now */
