@@ -2,32 +2,7 @@
  * Copyright (c) 2003 Niels Provos <provos@citi.umich.edu>
  * All rights reserved.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * <LICENSEHERE>
  */
 
 #include <sys/param.h>
@@ -108,6 +83,29 @@ condition_match_addr(const struct template *tmpl, const struct ip_hdr *ip,
 	return (1);
 }
 
+/*
+ * Match the IP protocol of the packet
+ */
+
+int
+condition_match_proto(const struct template *tmpl, const struct ip_hdr *ip,
+    u_short iplen, void *arg)
+{
+	int *proto = arg;
+	return (ip->ip_p == *proto);
+}
+
+/*
+ * Match anything
+ */
+
+int
+condition_match_otherwise(const struct template *tmpl, const struct ip_hdr *ip,
+    u_short iplen, void *arg)
+{
+	return 1;
+}
+
 static int
 daysec(const struct tm *tm)
 {
