@@ -54,6 +54,7 @@
 #ifdef HAVE_TIME_H
 #include <time.h>
 #endif
+#include <fnmatch.h>
 
 #include <pcap.h>
 
@@ -73,7 +74,13 @@
 #include "arp.h"
 #include "pool.h"
 #include "dhcpclient.h"
+#include "util.h"
+#include "log.h"
 
+/* Tailq that holds all subsystems */
+struct subsystemqueue subsystems;
+
+/* Tree that contains all templates */
 struct templtree templates;
 
 /* Counter for addresses in 169.254/16 that we assign for DHCP */
@@ -108,17 +115,23 @@ SPLAY_GENERATE(porttree, port, node, port_compare);
 void
 config_init(void)
 {
+	TAILQ_INIT(&subsystems);
 	SPLAY_INIT(&templates);
+
+	no_spoof.new_src.addr_type = ADDR_TYPE_NONE;	/* default is no source spoofing... */
+	no_spoof.new_dst.addr_type = ADDR_TYPE_NONE;	/* ... and no destination spoofing */
 }
 
 void
 config_read(char *config)
 {
+	extern int honeyd_ignore_parse_errors;
 	FILE *fp;
 
 	if ((fp = fopen(config, "r")) == NULL)
 		err(1, "fopen(%s)", config);
-	if (parse_configuration(fp, config) == -1)
+	if (parse_configuration(fp, config) == -1 &&
+	    !honeyd_ignore_parse_errors)
 		errx(1, "parsing configuration file failed");
 
 	fclose(fp);
@@ -133,6 +146,43 @@ template_find(const char *name)
 	return (SPLAY_FIND(templtree, &templates, &tmp));
 }
 
+void
+template_list_glob(struct evbuffer *buffer, const char *pattern)
+{
+	struct template *tmpl, *last = NULL;
+	struct evbuffer *tmp = NULL;
+	int count = 0;
+
+	tmpl = template_find(pattern);
+	if (tmpl != NULL) {
+		template_print(buffer, tmpl);
+		return;
+	}	
+
+	if ((tmp = evbuffer_new()) == NULL)
+		err(1, "%s: malloc");
+
+	SPLAY_FOREACH(tmpl, templtree, &templates) {
+		/* Ignore it if it does not match */
+		if (fnmatch(pattern, tmpl->name, 0))
+			continue;
+		count++;
+		evbuffer_add_printf(tmp, "%4d. %s (%s)\n",
+		    count,
+		    tmpl->name,
+		    tmpl->person != NULL ? tmpl->person->name : "undefined");
+
+		last = tmpl;
+	}
+
+	if (count == 1) {
+		template_print(buffer, last);
+	} else {
+		evbuffer_add_buffer(buffer, tmp);
+	}
+	evbuffer_free(tmp);
+}
+
 /*
  * Checks if condition for each template in the list is matched.
  * Return the first template that we match.
@@ -198,6 +248,9 @@ template_create(const char *name)
 	TAILQ_INIT(&tmpl->subsystems);
 	TAILQ_INIT(&tmpl->dynamic);
 
+	/* No spoofing, by default */
+	tmpl->spoof = no_spoof;
+
 	/* Crank ref counter */
 	tmpl->refcnt++;
 
@@ -349,11 +402,39 @@ port_action_clone(struct action *dst, const struct action *src)
 	}
 }
 
+void
+port_encapsulation_free(struct port_encapsulate *tmp)
+{
+	struct port *port = tmp->port;
+
+	TAILQ_REMOVE(&port->pending, tmp, next);
+	event_del(&tmp->ev);
+	
+	/* Remove the reference to the pending connection */
+	if (tmp->hdr != NULL)
+		tmp->hdr->pending = NULL;
+
+	free(tmp);
+}
+
 void
 port_free(struct template *tmpl, struct port *port)
 {
+	struct port_encapsulate *tmp;
+
+	/* Remove pending connections */
+	while ((tmp = TAILQ_FIRST(&port->pending)) != NULL) {
+		/* This might not be the correct way to clean this up */
+		if (tmp->hdr != NULL && tmp->hdr->type == SOCK_STREAM)
+			tcp_connectfail(tmp->con);
+	
+		port_encapsulation_free(tmp);
+	}
+
 	SPLAY_REMOVE(porttree, &tmpl->ports, port);
 
+	
+
 	if (port->sub_conport != NULL) {
 		/* Back pointer to connection object.
 		 * It allows us to remove the reference to this object
@@ -367,7 +448,7 @@ port_free(struct template *tmpl, struct port *port)
 	if (port->sub != NULL)
 		TAILQ_REMOVE(&port->sub->ports, port, next);
 	if (port->sub_fd != -1)
-		close(port->sub_fd);
+		fdshare_close(port->sub_fd);
 	if (port->action.action != NULL)
 		free (port->action.action);
 	if (port->action.aitop != NULL)
@@ -390,6 +471,7 @@ port_insert(struct template *tmpl, int proto, int number,
 	if ((port = calloc(1, sizeof(struct port))) == NULL)
 		err(1, "%s: calloc", __func__);
 
+	TAILQ_INIT(&port->pending);
 	port->sub = NULL;
 	port->sub_fd = -1;
 	port->proto = proto;
@@ -502,7 +584,6 @@ template_clone(const char *newname, const struct template *tmpl,
 	struct template *newtmpl;
 	struct port *port;
 	struct addr addr;
-	char *argv[4];
 	int isipaddr = 0;
 
 	if ((newtmpl = template_create(newname)) == NULL)
@@ -552,6 +633,7 @@ template_clone(const char *newname, const struct template *tmpl,
 	newtmpl->drop_inrate = tmpl->drop_inrate;
 	newtmpl->drop_synrate = tmpl->drop_synrate;
 	newtmpl->flags = tmpl->flags;
+	newtmpl->spoof = tmpl->spoof;
 
 	/* We need to remove this when cloning */
 	newtmpl->flags &= ~TEMPLATE_DYNAMIC_CHILD;
@@ -573,17 +655,13 @@ template_clone(const char *newname, const struct template *tmpl,
 		 * subsystem has not been specified as shared in the
 		 * configuration.
 		 */
-		if (!sub->shared) {
-			template_subsystem(newtmpl, sub->cmdstring, start, 0);
+		if (!(sub->flags & SUBSYSTEM_SHARED)) {
+			template_subsystem(newtmpl, sub->cmdstring,sub->flags);
 			continue;
 		}
 
 		/* Otherwise, just create references */
-		TAILQ_INSERT_TAIL(&sub->templates, template_ref(newtmpl),next);
-
-		if (isipaddr)
-			SPLAY_INSERT(subtmpltree, &sub->root,
-			    template_ref(newtmpl));
+		subsystem_insert_template(sub, newtmpl);
 
 		template_insert_subsystem(newtmpl, sub);
 	}
@@ -592,29 +670,16 @@ template_clone(const char *newname, const struct template *tmpl,
 	if (!start)
 		return (newtmpl);
 
-	argv[0] = "/bin/sh";
-	argv[1] = "-c";
-	argv[3] = NULL;
-
 	/* Start background processes */
 	TAILQ_FOREACH(container, &newtmpl->subsystems, next) {
-		struct subsystem *sub = container->sub;
-
-		/* Has the subsystem started already? */
-		if (sub->cmd.pid != -1)
-			continue;
-
-		argv[2] = sub->cmdstring;
-		if (cmd_subsystem(newtmpl, sub, "/bin/sh", argv) == -1)
-			errx(1, "%s: can not start subsystem \"%s\" for %s",
-			    sub->cmdstring, newtmpl->name);
+		template_subsystem_start(newtmpl, container->sub);
 	}
 
 	return (newtmpl);
 }
 
 int
-template_subsystem(struct template *tmpl, char *subsystem, int isaddr, int shared)
+template_subsystem(struct template *tmpl, char *subsystem, int flags)
 {
 	struct subsystem *sub;
 	
@@ -629,26 +694,22 @@ template_subsystem(struct template *tmpl, char *subsystem, int isaddr, int share
 	SPLAY_INIT(&sub->root);
 	TAILQ_INIT(&sub->templates);
 
-	TAILQ_INSERT_TAIL(&sub->templates, template_ref(tmpl), next);
+	subsystem_insert_template(sub, tmpl);
 
-	/*
-	 * Only make this template available to the subsystem if it is
-	 * an internet address.
-	 */
-	if (isaddr)
-		SPLAY_INSERT(subtmpltree, &sub->root, template_ref(tmpl));
 	sub->cmd.pid = -1;
-	sub->shared = shared;
+	sub->flags |= flags;
 
 	template_insert_subsystem(tmpl, sub);
 
+	/* Remember this subsystem for lookup through the UI */
+	TAILQ_INSERT_TAIL(&subsystems, sub, next);
+
 	return (0);
 }
 
 void
-template_subsystem_free(struct subsystem *sub)
+template_subsystem_free_ports(struct subsystem *sub)
 {
-	struct template *tmpl;
 	struct port *port;
 
 	for (port = TAILQ_FIRST(&sub->ports); port;
@@ -657,17 +718,84 @@ template_subsystem_free(struct subsystem *sub)
 		/* Free all ports for this subsystem */
 		port_free(port->subtmpl, port);
 	}
+}
+
+struct subsystem *
+template_subsystem_find(const char *name)
+{
+	struct subsystem *sub;
+
+	TAILQ_FOREACH(sub, &subsystems, next) {
+		if (!strcmp(name, sub->cmdstring))
+			return (sub);
+	}
+
+	return (NULL);
+}
+
+void
+template_subsystem_list_glob(struct evbuffer *buffer, const char *pattern)
+{
+	struct subsystem *sub, *last = NULL;
+	struct evbuffer *tmp = NULL;
+	int count = 0;
+
+	sub = template_subsystem_find(pattern);
+	if (sub != NULL) {
+		subsystem_print(buffer, sub);
+		return;
+	}
+
+	if ((tmp = evbuffer_new()) == NULL)
+		err(1, "%s: malloc");
+
+	TAILQ_FOREACH(sub, &subsystems, next) {
+		/* Ignore it if it does not match */
+		if (fnmatch(pattern, sub->cmdstring, 0))
+			continue;
+
+		count++;
+		evbuffer_add_printf(tmp, "%4d. %s (%d)\n",
+		    count, sub->cmdstring, sub->cmd.pid);
+
+		last = sub;
+	}
+
+	if (count == 1) {
+		subsystem_print(buffer, last);
+	} else {
+		evbuffer_add_buffer(buffer, tmp);
+	}
+	evbuffer_free(tmp);
+}
+
+void
+template_subsystem_free(struct subsystem *sub)
+{
+	struct template_container *cont;
+	struct template *tmpl;
+
+	TAILQ_REMOVE(&subsystems, sub, next);
+
+	template_subsystem_free_ports(sub);
 		
 	/* 
 	 * As we are removing all templates for this subsystem, we
 	 * actually do not need to remove them from the Splay.
 	 */
-	TAILQ_FOREACH(tmpl, &sub->templates, next) {
+	for (cont = TAILQ_FIRST(&sub->templates); cont;
+	    cont = TAILQ_FIRST(&sub->templates)) {
+		TAILQ_REMOVE(&sub->templates, cont, next);
+		tmpl = cont->tmpl;
+		free(cont);
+
 		template_remove_subsystem(tmpl, sub);
 
 		template_free(tmpl);
 	}
 
+	cmd_free(&sub->cmd);
+
 	free(sub->cmdstring);
 	free(sub);
 }
@@ -691,7 +819,8 @@ template_insert_dynamic(struct template *tmpl, struct template *child,
 		*cond = *condition;
 
 	/* Con up a new template name */
-	snprintf(newname, sizeof(newname), "%s_%s", tmpl->name, child->name);
+	snprintf(newname, sizeof(newname), "%s_%d_%s",
+	    tmpl->name, tmpl->dynamic_rulenr++, child->name);
 	if ((cond->tmpl = template_clone(newname, child, NULL, 0)) == NULL) {
 		fprintf(stderr, "Failed to clone %s from %s\n",
 		    newname, child->name);
@@ -728,7 +857,109 @@ template_get_dhcp_address(struct addr *addr)
 	return (addr_aton(address, addr));
 }
 
-/* Everything is unittest related below this */
+void
+template_subsystem_start(struct template *tmpl, struct subsystem *sub)
+{
+	extern int honeyd_verify_config;
+	char *p;
+	char *argv[4];
+	char line[512];	/* for command replacement */
+	char *name = tmpl->name;
+	struct addr addr;
+	int isipaddr = addr_aton(name, &addr) != -1;
+
+
+	/* Has the subsystem started already? */
+	if (sub->cmd.pid != -1 || honeyd_verify_config)
+		return;
+
+	gettimeofday(&sub->tv_restart, NULL);
+
+	argv[0] = "/bin/sh";
+	argv[1] = "-c";
+	argv[3] = NULL;
+
+	strlcpy(line, sub->cmdstring, sizeof(line));
+	if (isipaddr) {
+		/*
+		 * Provide a subsystem with it's own IP address.
+		 * For a shared subsystem, this is going to fail.
+		 */
+		while (strrpl(line, sizeof(line), "$ipsrc", name) != NULL)
+			;
+	}
+	p = honeyd_logdate();
+        while (strrpl(line, sizeof(line), "$date", p) != NULL)
+                ;
+
+	argv[2] = line;
+	if (cmd_subsystem(tmpl, sub, "/bin/sh", argv) == -1)
+		errx(1, "%s: can not start subsystem \"%s\" for %s",
+		    sub->cmdstring, name);
+
+}
+
+void
+template_print(struct evbuffer *buffer, struct template *tmpl)
+{
+	struct port *port;
+
+	evbuffer_add_printf(buffer, "template %s:\n", tmpl->name);
+	evbuffer_add_printf(buffer, "  personality: %s\n",
+	    tmpl->person != NULL ? tmpl->person->name : "undefined");
+	if (tmpl->ethernet_addr != NULL)
+		evbuffer_add_printf(buffer, "  ethernet address: %s\n",
+		    addr_ntoa(tmpl->ethernet_addr));
+	evbuffer_add_printf(buffer, "  IP id: %u\n", tmpl->id);
+	evbuffer_add_printf(buffer, "  TCP seq: %lx\n", tmpl->seq);
+	evbuffer_add_printf(buffer, "  TCP drop: in: %d syn: %d\n",
+	    tmpl->drop_inrate, tmpl->drop_synrate);
+	evbuffer_add_printf(buffer, "  refcnt: %d\n", tmpl->refcnt);
+	evbuffer_add_printf(buffer, "  ports:\n");
+
+	SPLAY_FOREACH(port, porttree, &tmpl->ports) {
+		char *type;
+		switch (port->action.status) {
+		case PORT_OPEN:
+			type = "open";
+			break;
+		case PORT_PROXY:
+			type = "proxy";
+			break;
+		case PORT_BLOCK:
+			type = "block";
+			break;
+		case PORT_RESET:
+			type = "reset";
+			break;
+		case PORT_SUBSYSTEM:
+			type = "subsystem";
+			break;
+		case PORT_PYTHON:
+			type = "python";
+			break;
+		default:
+			type = "reserved";
+			break;
+		}
+		evbuffer_add_printf(buffer, "    %s %5d %s",
+		    port->proto == IP_PROTO_TCP ? "tcp" : "udp",
+		    port->number, type);
+		evbuffer_add_printf(buffer, "\n");
+		if (port->action.status == PORT_SUBSYSTEM) {
+			evbuffer_add_printf(buffer, "\t%s\n",
+			    port->sub->cmdstring);
+		} else if (port->action.status == PORT_OPEN &&
+		    port->action.action != NULL) {
+			evbuffer_add_printf(buffer, "\t%s\n",
+			    port->action.action);
+		}
+	}
+}
+
+/***************************************************************************
+ * Everything is unittest related below this
+ ***************************************************************************/
 
 void
 template_delay_cb(int fd, short which, void *arg)
@@ -803,6 +1034,8 @@ void
 template_test_measure(int count)
 {
 	extern rand_t *honeyd_rand;
+	extern void
+	  honeyd_recv_cb(u_char *, const struct pcap_pkthdr *, const u_char *);
 	u_char pkt[1500];
 	struct interface inter;
 	struct pcap_pkthdr pkthdr;
@@ -838,7 +1071,7 @@ template_test_measure(int count)
 		    IP_PROTO_TCP, src.addr_ip, dst.addr_ip);
 		ip_checksum(pkt, iplen);
 
-		honeyd_recv_cb(&inter, &pkthdr, pkt);
+		honeyd_recv_cb((u_char *)&inter, &pkthdr, pkt);
 	}
 	gettimeofday(&tv_end, NULL);
 	timersub(&tv_end, &tv_start, &tv_end);
@@ -885,7 +1118,6 @@ template_packet_test(void)
 	fprintf(stderr, "\t%s: OK\n", __func__);
 }
 
-
 void
 template_test(void)
 {
