@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2002, 2003, 2004 Niels Provos <provos@citi.umich.edu>
+ * Copyright (c) 2002, 2003, 2004, 2005 Niels Provos <provos@citi.umich.edu>
  * All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
@@ -63,11 +63,13 @@
 #endif
 #include <unistd.h>
 #include <getopt.h>
-#include <dnet.h>
+#include <pwd.h>
+#include <assert.h>
 
 #undef timeout_pending
 #undef timeout_initialized
 
+#include <dnet.h>
 #include <event.h>
 
 #include "honeyd.h"
@@ -98,8 +100,10 @@
 #include "rrdtool.h"
 #include "histogram.h"
 #include "update.h"
+#include "util.h"
 
 #ifdef HAVE_PYTHON
+#include <Python.h>
 #include "pyextend.h"
 #endif
 
@@ -112,13 +116,15 @@ enum forward honeyd_route_packet(struct ip_hdr *, u_int, struct addr *,
 
 void tcp_retrans_timeout(int, short, void *);
 void icmp_error_send(struct template *, struct addr *, uint8_t, uint8_t,
-    struct ip_hdr *);
+    struct ip_hdr *, struct spoof);
 
 struct tree tcpcons;
 struct conlru tcplru;
 struct tree udpcons;
 struct conlru udplru;
 
+struct spoof no_spoof;	/* spoof settings for default packet processing */
+
 struct config config = {
 	NULL,
 	PATH_HONEYDDATA "/nmap.prints",
@@ -151,6 +157,7 @@ int			 honeyd_nchildren;
 int			 honeyd_ttl = HONEYD_DFL_TTL;
 struct tcp_con		 honeyd_tmp;
 int                      honeyd_show_include_dir;
+int                      honeyd_show_data_dir;
 int                      honeyd_show_version;
 int                      honeyd_show_usage;
 int			 honeyd_debug;
@@ -159,7 +166,10 @@ gid_t			 honeyd_gid = 32767;
 int			 honeyd_needsroot;	/* Need different IDs */
 int			 honeyd_disable_webserver = 0;
 int			 honeyd_disable_update = 0;
+int			 honeyd_ignore_parse_errors = 0;
+int			 honeyd_verify_config = 0;
 int			 honeyd_webserver_fix_permissions = 0;
+char			*honeyd_webserver_address = "127.0.0.1";
 int			 honeyd_webserver_port = 80;
 char			*honeyd_webserver_root = PATH_HONEYDDATA \
 						"/webserver/htdocs";
@@ -173,13 +183,17 @@ static char		*servicelog = NULL;
 
 static struct option honeyd_long_opts[] = {
 	{"include-dir", 0, &honeyd_show_include_dir, 1},
+	{"data-dir",    0, &honeyd_show_data_dir, 1},
 	{"version",     0, &honeyd_show_version, 1},
 	{"help",        0, &honeyd_show_usage, 1},
+	{"webserver-address", required_argument, NULL, 'A'},
 	{"webserver-port", required_argument, NULL, 'W'},
 	{"webserver-root", required_argument, NULL, 'X'},
 	{"rrdtool-path", required_argument, NULL, 'Y'},
 	{"disable-webserver", 0, &honeyd_disable_webserver, 1},
 	{"disable-update", 0, &honeyd_disable_update, 1},
+	{"verify-config", 0, &honeyd_verify_config, 1},
+	{"ignore-parse-errors", 0, &honeyd_ignore_parse_errors, 1},
 	{"fix-webserver-permissions", 0, &honeyd_webserver_fix_permissions, 1},
 	{0, 0, 0, 0}
 };
@@ -203,21 +217,40 @@ usage(void)
 	    "  -g gid		  Set the gid Honeyd should run as.\n"
 	    "  -f configfile          Read configuration from file.\n"
 	    "  -c host:port:name:pass Reports starts to collector.\n"
+	    "  --webserver-address=address Address on which webserver listens.\n"
 	    "  --webserver-port=port  Port on which webserver listens.\n"
 	    "  --webserver-root=path  Root of document tree.\n"
 	    "  --fix-webserver-permissions Change ownership and permissions.\n"
 	    "  --rrdtool-path=path    Path to rrdtool.\n"
 	    "  --disable-webserver    Disables internal webserver\n"
 	    "  --disable-update       Disables checking for security fixes.\n"
+	    "  --verify-config        Verify configuration file then exit.\n"
 	    "  -V, --version          Print program version and exit.\n"
 	    "  -h, --help             Print this message and exit.\n"
 	    "\n"
 	    "For plugin development:\n"
-	    "  --include-dir          Prints out header files directory and exits.\n");
+	    "  --include-dir          Prints out header files directory and exits.\n"
+	    "  --data-dir             Prints out data/plug-in directory and exits.\n");
 	
 	exit(1);
 }
 
+/* XXX ches debug */
+void
+print_spoof(char *msg, struct spoof s) {
+	char buf2[100], buf3[100];
+
+	if (s.new_src.addr_type == ADDR_TYPE_NONE)
+		strlcpy(buf2, "**no addr**", sizeof(buf2));
+	else
+		addr_ntop(&s.new_src, buf2, sizeof(buf2));
+
+	if (s.new_dst.addr_type == ADDR_TYPE_NONE)
+		strlcpy(buf3, "**no addr**", sizeof(buf3));
+	else
+		addr_ntop(&s.new_dst, buf3, sizeof(buf3));
+}
+
 void
 honeyd_settcp(struct tcp_con *con, struct ip_hdr *ip, struct tcp_hdr *tcp,
     int local)
@@ -394,10 +427,17 @@ void
 honeyd_init(void)
 {
 	struct rlimit rl;
+	struct passwd *pwd;
 
 	/* Record our start time */
 	gettimeofday(&honeyd_uptime, NULL);
 
+	/* Find the correct ids for nobody */
+	if ((pwd = getpwnam("nobody")) != NULL) {
+		honeyd_uid = pwd->pw_uid;
+		honeyd_gid = pwd->pw_gid;
+	}
+
 	/* Initalize ongoing connection state */
 	SPLAY_INIT(&tcpcons);
 	TAILQ_INIT(&tcplru);
@@ -608,13 +648,13 @@ honeyd_delay_cb(int fd, short which, void *arg)
 		ip->ip_ttl++;
 		ip_checksum(ip, ip->ip_hl << 2);
 		icmp_error_send(tmpl, &delay->src,
-		    ICMP_TIMEXCEED, ICMP_TIMEXCEED_INTRANS, ip);
+		    ICMP_TIMEXCEED, ICMP_TIMEXCEED_INTRANS, ip, delay->spoof);
 	} else if (delay->flags & DELAY_UNREACH) {
 		/* Fix up TTL */
 		ip->ip_ttl++;
 		ip_checksum(ip, ip->ip_hl << 2);
 		icmp_error_send(tmpl, &delay->src,
-		    ICMP_UNREACH, ICMP_UNREACH_NET, ip);
+		    ICMP_UNREACH, ICMP_UNREACH_NET, ip, delay->spoof);
 	} else if (delay->flags & DELAY_EXTERNAL) {
 		struct addr dst;
 		addr_pack(&dst, ADDR_TYPE_IP, IP_ADDR_BITS,
@@ -715,7 +755,8 @@ honeyd_delay_cb(int fd, short which, void *arg)
 
 void
 honeyd_delay_packet(struct template *tmpl, struct ip_hdr *ip, u_int iplen,
-    const struct addr *src, const struct addr *dst, int ms, int flags)
+    const struct addr *src, const struct addr *dst, int ms, int flags,
+    struct spoof spoof)
 {
 	struct delay *delay, tmp_delay;
 	struct timeval tv;
@@ -754,6 +795,7 @@ honeyd_delay_packet(struct template *tmpl, struct ip_hdr *ip, u_int iplen,
 		delay->dst = *dst;
 	delay->tmpl = template_ref(tmpl);
 	delay->flags = flags;
+	delay->spoof = spoof;
 
 	if (ms) {
 		evtimer_set(&delay->timeout, honeyd_delay_callback, delay);
@@ -773,7 +815,7 @@ honeyd_delay_packet(struct template *tmpl, struct ip_hdr *ip, u_int iplen,
  */
 
 void
-honeyd_ip_send(u_char *pkt, u_int iplen)
+honeyd_ip_send(u_char *pkt, u_int iplen, struct spoof spoof)
 {
 	struct template *tmpl = NULL;
 	struct ip_hdr *ip = (struct ip_hdr *)pkt;
@@ -781,13 +823,16 @@ honeyd_ip_send(u_char *pkt, u_int iplen)
 	int delay = 0, flags = 0;
 	struct addr addr, src;
 
+print_spoof("honeyd_ip_send", spoof);
+
 	if (iplen > HONEYD_MTU) {
 		u_short off = ntohs(ip->ip_off);
 		if ((off & IP_DF) == 0)
-			ip_send_fragments(HONEYD_MTU, ip, iplen);
+			ip_send_fragments(HONEYD_MTU, ip, iplen, spoof);
 		goto drop;
 	}
-
+	if (spoof.new_dst.addr_type != ADDR_TYPE_NONE)
+		ip->ip_dst = spoof.new_dst.addr_ip;
 	addr_pack(&addr, ADDR_TYPE_IP, IP_ADDR_BITS, &ip->ip_dst, IP_ADDR_LEN);
 	addr_pack(&src, ADDR_TYPE_IP, IP_ADDR_BITS, &ip->ip_src, IP_ADDR_LEN);
 
@@ -817,6 +862,8 @@ honeyd_ip_send(u_char *pkt, u_int iplen)
 		if (addr_cmp(&src, &router->addr) == 0)
 			ip->ip_ttl++; /* XXX - Ugly hack */
 
+		if (spoof.new_src.addr_type != ADDR_TYPE_NONE)
+			ip->ip_src = spoof.new_src.addr_ip;
 		res = honeyd_route_packet(ip, iplen, &router->addr, &addr,
 		    &delay);
 		if (res == FW_DROP)
@@ -828,8 +875,13 @@ honeyd_ip_send(u_char *pkt, u_int iplen)
 	if (res == FW_EXTERNAL)
 		flags |= DELAY_EXTERNAL;
 	
+	if (spoof.new_src.addr_type != ADDR_TYPE_NONE)
+		ip->ip_src = spoof.new_src.addr_ip;
+	if (spoof.new_dst.addr_type != ADDR_TYPE_NONE)
+		ip->ip_dst = spoof.new_dst.addr_ip;
+
 	/* Delay the packet if necessary, otherwise deliver it directly */
-	honeyd_delay_packet(tmpl, ip, iplen, NULL, NULL, delay, flags);
+	honeyd_delay_packet(tmpl, ip, iplen, NULL, NULL, delay, flags, spoof);
 	return;
 
  drop:
@@ -898,6 +950,10 @@ void
 tcp_free(struct tcp_con *con)
 {
 	struct port *port = con->port;
+	struct port_encapsulate *pending = con->conhdr.pending;
+
+	if (pending != NULL)
+		port_encapsulation_free(pending);
 
 	if (port != NULL)
 		port_free(port->subtmpl, port);
@@ -988,6 +1044,10 @@ udp_free(struct udp_con *con)
 {
 	struct conbuffer *buf;
 	struct port *port = con->port;
+	struct port_encapsulate *pending = con->conhdr.pending;
+
+	if (pending != NULL)
+		port_encapsulation_free(pending);
 
 	if (port != NULL)
 		port_free(port->subtmpl, port);
@@ -1083,6 +1143,9 @@ honeyd_varexpand(struct tcp_con *con, char *line, u_int linesize)
 	p = addr_ntoa(&dst);
         while (strrpl(line, linesize, "$ipdst", p) != NULL)
                 ;
+	p = honeyd_logdate();
+        while (strrpl(line, linesize, "$date", p) != NULL)
+                ;
 	snprintf(asc, sizeof(asc), "%d", con->con_sport);
         while (strrpl(line, linesize, "$sport", asc) != NULL)
                 ;
@@ -1206,7 +1269,8 @@ generic_connect(struct template *tmpl, struct tuple *hdr,
 		goto out;
 
 	if ((port = port_find(tmpl, proto, hdr->dport)) == NULL) {
-		action = &tmpl->tcp;
+		/* We need to use the default action for the protocol */
+		action = proto == IP_PROTO_TCP ? &tmpl->tcp : &tmpl->udp;
 	} else
 		action = &port->action;
 
@@ -1220,7 +1284,7 @@ generic_connect(struct template *tmpl, struct tuple *hdr,
 
 	/* Connect to the already started sub system */
 	if (action->status == PORT_SUBSYSTEM) {
-		if (cmd_subsystem_connect(hdr, cmd, port, con) == -1)
+		if (cmd_subsystem_schedule_connect(hdr, cmd, port, con) == -1)
 			goto out;
 		return;
 	} else if (action->status == PORT_PYTHON) {
@@ -1240,6 +1304,7 @@ generic_connect(struct template *tmpl, struct tuple *hdr,
 	}
 
 	if (action->status == PORT_OPEN || action->aitop == NULL) {
+		assert(action->action != NULL && strlen(action->action));
 		strlcpy(line, action->action, sizeof(line));
 		honeyd_varexpand(con, line, sizeof(line));
 		/* Copy for print out */
@@ -1300,6 +1365,8 @@ tcp_send(struct tcp_con *con, uint8_t flags, u_char *payload, u_int len)
 	int dontfragment = 0;
 	char *options;
 	uint16_t id = rand_uint16(honeyd_rand);
+	struct spoof spoof;
+	struct template *tmpl = con->tmpl;
 
 	if (con->window)
 		window = con->window;
@@ -1360,6 +1427,11 @@ tcp_send(struct tcp_con *con, uint8_t flags, u_char *payload, u_int len)
 		tcp_personality_options(con, tcp, options);
 
 	iplen = IP_HDR_LEN + (tcp->th_off << 2) + len;
+
+	if (tmpl != NULL)
+		spoof = tmpl->spoof;
+	else
+		spoof = no_spoof;
 	
 	/* Src and Dst are reversed both for ip and tcp */
 	ip_pack_hdr(pkt, 0, iplen, id,
@@ -1371,7 +1443,7 @@ tcp_send(struct tcp_con *con, uint8_t flags, u_char *payload, u_int len)
 	hooks_dispatch(IP_PROTO_TCP, HD_OUTGOING, &con->conhdr,
 	    pkt, iplen);
 
-	honeyd_ip_send(pkt, iplen);
+	honeyd_ip_send(pkt, iplen, spoof);
 
 	return (len);
 }
@@ -1453,7 +1525,7 @@ tcp_sendfin(struct tcp_con *con)
 void
 icmp_send(struct template *tmpl,
     u_char *pkt, uint8_t tos, u_int iplen, uint16_t df, uint8_t ttl,
-    int proto, ip_addr_t src, ip_addr_t dst)
+    int proto, ip_addr_t src, ip_addr_t dst, struct spoof spoof)
 {
 	struct ip_hdr ip;
 	uint16_t ipid;
@@ -1472,12 +1544,12 @@ icmp_send(struct template *tmpl,
 	ip_pack_hdr(pkt, tos, iplen, ipid, df ? IP_DF: 0, ttl,
 	    IP_PROTO_ICMP, src, dst);
 
-	honeyd_ip_send(pkt, iplen);
+	honeyd_ip_send(pkt, iplen, spoof);
 }
 
 void
 icmp_error_send(struct template *tmpl, struct addr *addr,
-    uint8_t type, uint8_t code, struct ip_hdr *rip)
+    uint8_t type, uint8_t code, struct ip_hdr *rip, struct spoof spoof)
 {
 	u_char *pkt;
 	u_int iplen;
@@ -1499,7 +1571,7 @@ icmp_error_send(struct template *tmpl, struct addr *addr,
 
 	icmp_pack_hdr_quote(pkt + IP_HDR_LEN, type, code, 0, rip, quotelen);
 	icmp_send(tmpl, pkt, tos, iplen, df ? IP_DF: 0, ttl,
-	    IP_PROTO_ICMP, addr->addr_ip, rip->ip_src);
+	    IP_PROTO_ICMP, addr->addr_ip, rip->ip_src, spoof);
 }
 
 /*
@@ -1536,7 +1608,7 @@ icmp_error_send(struct template *tmpl, struct addr *addr,
 void
 icmp_echo_reply(struct template *tmpl,
     struct ip_hdr *rip, uint8_t code, uint8_t tos,
-    uint16_t offset, uint8_t ttl, u_char *payload, u_int len)
+    uint16_t offset, uint8_t ttl, u_char *payload, u_int len, struct spoof spoof)
 {
 	u_char *pkt;
 	u_int iplen;
@@ -1556,7 +1628,7 @@ icmp_echo_reply(struct template *tmpl,
 		payload, len);
 
 	icmp_send(tmpl, pkt, tos, iplen, offset, ttl,
-	    IP_PROTO_ICMP, rip->ip_dst, rip->ip_src);
+	    IP_PROTO_ICMP, rip->ip_dst, rip->ip_src, spoof);
 }
 
 /*
@@ -1578,7 +1650,7 @@ icmp_echo_reply(struct template *tmpl,
  */ 
 void
 icmp_timestamp_reply(struct template *tmpl, struct ip_hdr *rip,
-    struct icmp_msg_timestamp* icmp_rip, uint8_t ttl)
+    struct icmp_msg_timestamp* icmp_rip, uint8_t ttl, struct spoof spoof)
 {
 	u_char *pkt;
 	u_int iplen;
@@ -1614,7 +1686,7 @@ icmp_timestamp_reply(struct template *tmpl, struct ip_hdr *rip,
 	
 	memcpy(pkt + IP_HDR_LEN, &icmp_time, sizeof(icmp_time));
 	icmp_send(tmpl, pkt, rip->ip_tos, iplen, rip->ip_off, ttl,
-	    IP_PROTO_ICMP, rip->ip_dst, rip->ip_src);
+	    IP_PROTO_ICMP, rip->ip_dst, rip->ip_src, spoof);
 }
 
 /*
@@ -1637,7 +1709,7 @@ icmp_timestamp_reply(struct template *tmpl, struct ip_hdr *rip,
  */ 
 void
 icmp_mask_reply(struct template *tmpl, struct ip_hdr *rip, 
-	struct icmp_msg_idseq *idseq, uint8_t ttl, uint32_t addrmask)
+	struct icmp_msg_idseq *idseq, uint8_t ttl, uint32_t addrmask, struct spoof spoof)
 {
 	u_char *pkt;
 	u_int iplen;
@@ -1656,7 +1728,7 @@ icmp_mask_reply(struct template *tmpl, struct ip_hdr *rip,
 	
 	memcpy(pkt + IP_HDR_LEN, &mask, sizeof(mask));
 	icmp_send(tmpl, pkt, rip->ip_tos, iplen, rip->ip_off, ttl,
-	    IP_PROTO_ICMP, rip->ip_dst, rip->ip_src);
+	    IP_PROTO_ICMP, rip->ip_dst, rip->ip_src, spoof);
 }
 
 /*
@@ -1678,7 +1750,7 @@ icmp_mask_reply(struct template *tmpl, struct ip_hdr *rip,
  */ 
 void
 icmp_info_reply(struct template *tmpl, struct ip_hdr *rip, 
-		struct icmp_msg_idseq *idseq, uint8_t ttl)
+		struct icmp_msg_idseq *idseq, uint8_t ttl, struct spoof spoof)
 {
 	u_char *pkt;
 	u_int iplen;
@@ -1696,7 +1768,7 @@ icmp_info_reply(struct template *tmpl, struct ip_hdr *rip,
 	
 	memcpy(pkt + IP_HDR_LEN, &inforeply, sizeof(inforeply));
 	icmp_send(tmpl, pkt, rip->ip_tos, iplen, rip->ip_off, ttl,
-	    IP_PROTO_ICMP, rip->ip_dst, rip->ip_src);
+	    IP_PROTO_ICMP, rip->ip_dst, rip->ip_src, spoof);
 }
 
 void
@@ -1850,8 +1922,15 @@ tcp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 	if (pktlen < (ip->ip_hl << 2) + TCP_HDR_LEN)
 		return;
 
+	/* 
+	 * Check if we have a real connection header for this connection, so
+	 * that we can look at potential flags like local origination.
+	 */
 	honeyd_settcp(&honeyd_tmp, ip, tcp, 0);
-	hooks_dispatch(ip->ip_p, HD_INCOMING, &honeyd_tmp.conhdr, pkt, pktlen);
+	con = (struct tcp_con *)SPLAY_FIND(tree, &tcpcons, &honeyd_tmp.conhdr);
+
+	hooks_dispatch(ip->ip_p, HD_INCOMING, 
+	    con != NULL ? &con->conhdr : &honeyd_tmp.conhdr, pkt, pktlen);
 	
 	if (honeyd_block(tmpl, IP_PROTO_TCP, ntohs(tcp->th_dport)))
 		goto justlog;
@@ -1868,7 +1947,6 @@ tcp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 
 	tiflags = tcp->th_flags;
 
-	con = (struct tcp_con *)SPLAY_FIND(tree, &tcpcons, &honeyd_tmp.conhdr);
 	if (con == NULL) {
 		if (honeyd_tmp.state != TCP_STATE_LISTEN)
 			goto kill;
@@ -1948,6 +2026,22 @@ tcp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 		return;
 	}
 
+	/*
+	 * Subsystems can remove ports on the fly - even when parts of the
+	 * 3-way handshake are in progress already.
+	 */
+
+	if (action != NULL) {
+		switch (action->status) {
+		case PORT_RESET:
+			goto dropwithreset;
+		case PORT_BLOCK:
+			goto drop;
+		default:
+			break;
+		}
+	}
+
 	th_seq = ntohl(tcp->th_seq);
 	th_ack = ntohl(tcp->th_ack);
 
@@ -2018,10 +2112,25 @@ tcp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 		connection_update(&tcplru, &con->conhdr);
 
 		if (tiflags & TH_FIN && !(con->flags & TCP_TARPIT)) {
-			if (con->cmd_pfd > 0)
-				shutdown(con->cmd_pfd, SHUT_WR);
-			else
+			if (con->cmd_pfd > 0) {
+				if (con->rlen == 0) {
+					/*
+					 * If we already transmitted all data,
+					 * we can completely shutdown the write
+					 * part of the connection.
+					 */
+					shutdown(con->cmd_pfd, SHUT_WR);
+				} else {
+					/* 
+					 * If we still have data to write to
+					 * our child process, we need to delay
+					 * the shutdown until later.
+					 */
+					con->cmd.fdgotfin = 1;
+				}
+			} else {
 				con->sentfin = 1;
+			}
 			con->state = TCP_STATE_CLOSE_WAIT;
 			dlen++;
 		}
@@ -2175,11 +2284,13 @@ udp_send(struct udp_con *con, u_char *payload, u_int len)
 	u_int iplen;
 	uint16_t id = rand_uint16(honeyd_rand);
 	int dontfragment = 0;
+	struct spoof spoof;
+	struct template *tmpl = con->tmpl;
 
 	/* Statistics */
 	con->conhdr.sent += len;
 
-	ip_personality(con->tmpl, &id);
+	ip_personality(tmpl, &id);
 
 	pkt = pool_alloc(pool_pkt);
 
@@ -2195,12 +2306,19 @@ udp_send(struct udp_con *con, u_char *payload, u_int len)
 
 	memcpy(pkt + IP_HDR_LEN + UDP_HDR_LEN, payload, len);
 
+ 	if (tmpl)
+ 		spoof = tmpl->spoof;
+ 	else
+ 		spoof = no_spoof;
+
 	ip_checksum(pkt, iplen);
 	
 	hooks_dispatch(IP_PROTO_UDP, HD_OUTGOING, &con->conhdr,
 	    pkt, iplen);
 
-	honeyd_ip_send(pkt, iplen);
+	honeyd_ip_send(pkt, iplen, spoof);
+
+	generic_timeout(&con->conhdr.timeout, HONEYD_UDP_WAIT);
 
 	return (len);
 }
@@ -2212,6 +2330,7 @@ udp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 	struct udp_hdr *udp;
 	struct udp_con *con, honeyd_udp;
 	struct addr addr;
+	struct spoof spoof;
 	
 	uint16_t uh_sum;
 	u_char *data;
@@ -2224,8 +2343,15 @@ udp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 	if (pktlen < (ip->ip_hl << 2) + UDP_HDR_LEN)
 		return;
 
+	/* 
+	 * Check if we have a real connection header for this connection, so
+	 * that we can look at potential flags like local origination.
+	 */
 	honeyd_setudp(&honeyd_udp, ip, udp, 0);
-	hooks_dispatch(ip->ip_p, HD_INCOMING, &honeyd_udp.conhdr, pkt, pktlen);
+	con = (struct udp_con *)SPLAY_FIND(tree, &udpcons, &honeyd_udp.conhdr);
+
+	hooks_dispatch(ip->ip_p, HD_INCOMING,
+	    con != NULL ? &con->conhdr : &honeyd_udp.conhdr, pkt, pktlen);
 
 	data = (u_char *)(pkt + (ip->ip_hl*4) + UDP_HDR_LEN);
 	dlen = ntohs(ip->ip_len) - (ip->ip_hl << 2) - UDP_HDR_LEN;
@@ -2242,8 +2368,6 @@ udp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 	if ((uh_sum && uh_sum != udp->uh_sum))
 		goto justlog;
 
-	con = (struct udp_con *)SPLAY_FIND(tree, &udpcons, &honeyd_udp.conhdr);
-
 	if (con == NULL) {
 		struct action *action;
 		action = honeyd_port(tmpl, IP_PROTO_UDP, portnum);
@@ -2283,7 +2407,18 @@ udp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 	    pktlen, 0, NULL);
 
 	addr_pack(&addr, ADDR_TYPE_IP, IP_ADDR_BITS, &ip->ip_dst, IP_ADDR_LEN);
-	icmp_error_send(tmpl, &addr, ICMP_UNREACH, ICMP_UNREACH_PORT, ip); 
+
+	/*
+	 * compute possible spoofed source for this error response to a UDP packet
+	 */
+	if (tmpl)
+		spoof = tmpl->spoof;
+	else
+		spoof = no_spoof;
+	// compute_spoof(&spoof, tmpl, &tmpl->spoof, ip->ip_src, ip->ip_dst);
+print_spoof("udp_recv_cb after", spoof);
+
+	icmp_error_send(tmpl, &addr, ICMP_UNREACH, ICMP_UNREACH_PORT, ip, spoof); 
 	return;
 
  justlog:
@@ -2308,7 +2443,10 @@ icmp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 	struct xp_fingerprint *xp_print = NULL;  /* JVR */
 	struct tuple icmphdr;
 	struct addr src, dst;
-	char asrc[20], adst[20];
+	struct spoof spoof;
+	char asrc[100], adst[100];
+	char osrc[100], odst[100];
+	char ssrc[100], sdst[100];
 	u_char *dat;
 	uint16_t cksum;
 	int dlen;
@@ -2354,11 +2492,27 @@ icmp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 	if (dlen < 0)
 		return;
 
-	/* AscII representation of addresses */
+	/* AscII representation of original addresses */
 	addr_pack(&src, ADDR_TYPE_IP, IP_ADDR_BITS, &ip->ip_src, IP_ADDR_LEN);
 	addr_pack(&dst, ADDR_TYPE_IP, IP_ADDR_BITS, &ip->ip_dst, IP_ADDR_LEN);
-	addr_ntop(&src, asrc, sizeof(asrc));
-	addr_ntop(&dst, adst, sizeof(adst));
+	addr_ntop(&src, osrc, sizeof(osrc));
+	addr_ntop(&dst, odst, sizeof(odst));
+
+	if (tmpl)
+		spoof = tmpl->spoof;
+	else
+		spoof = no_spoof;
+	addr_ntop(&spoof.new_src, ssrc, sizeof(ssrc));
+	addr_ntop(&spoof.new_dst, sdst, sizeof(sdst));
+
+	if (spoof.new_src.addr_type != ADDR_TYPE_NONE)
+		snprintf(adst, sizeof(adst), "%s (was %s)", ssrc, odst);
+	else
+		snprintf(adst, sizeof(adst), "%s", odst);
+	if (spoof.new_dst.addr_type != ADDR_TYPE_NONE)
+		snprintf(asrc, sizeof(asrc), "%s (was %s)", sdst, osrc);
+	else
+		snprintf(asrc, sizeof(asrc), "%s", osrc);
 
 	switch (icmp->icmp_type) {
 	case ICMP_ECHO:
@@ -2373,10 +2527,10 @@ icmp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 			    xp_print->flags.icmp_echo_tos_bits ? ip->ip_tos : 0,
 			    xp_print->flags.icmp_echo_df_bit ? IP_DF : 0,
 			    xp_print->ttl_vals.icmp_echo_reply_ttl.ttl_val,
-			    dat, dlen);
+			    dat, dlen, spoof);
 		} else
 			icmp_echo_reply(tmpl, ip,
-			    ICMP_CODE_NONE, 0, 0, honeyd_ttl, dat, dlen);
+			    ICMP_CODE_NONE, 0, 0, honeyd_ttl, dat, dlen, spoof);
 		break;
 
 	case ICMP_UNREACH:
@@ -2420,7 +2574,7 @@ icmp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 			    adst, asrc);
 		    
 			icmp_timestamp_reply(tmpl, ip, icmp_tstamp,
-			    xp_print->ttl_vals.icmp_timestamp_reply_ttl.ttl_val);
+			    xp_print->ttl_vals.icmp_timestamp_reply_ttl.ttl_val, spoof);
 		}
 	        break;
 
@@ -2435,7 +2589,7 @@ icmp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 		
 			icmp_mask_reply(tmpl, ip, icmp_idseq,
 			    xp_print->ttl_vals.icmp_addrmask_reply_ttl.ttl_val,
-			    HONEYD_ADDR_MASK);
+			    HONEYD_ADDR_MASK, spoof);
 		}
 		break;
 
@@ -2449,7 +2603,7 @@ icmp_recv_cb(struct template *tmpl, u_char *pkt, u_short pktlen)
 			    adst, asrc);
 		
 			icmp_info_reply(tmpl, ip, icmp_idseq,
-			    xp_print->ttl_vals.icmp_info_reply_ttl.ttl_val);
+			    xp_print->ttl_vals.icmp_info_reply_ttl.ttl_val, spoof);
 		}
 		break;
 
@@ -2616,6 +2770,9 @@ honeyd_route_packet(struct ip_hdr *ip, u_int iplen,
 
 	/* Send ICMP_TIMEXCEED from router address */
 	if (!ip->ip_ttl) {
+ 		struct spoof spoof = no_spoof;
+ 		spoof.new_src = host;
+ 
 		syslog(LOG_DEBUG, "TTL exceeded for dst %s at gw %s",
 		    addr_ntoa(addr), addr_ntoa(&host));
 
@@ -2624,7 +2781,7 @@ honeyd_route_packet(struct ip_hdr *ip, u_int iplen,
 		 * send the ICMP error message.
 		 */
 		tmpl = template_find_best(addr_ntoa(&host), ip, iplen);
-		honeyd_delay_packet(tmpl, ip, iplen, &host, NULL, delay, 0);
+		honeyd_delay_packet(tmpl, ip, iplen, &host, NULL, delay, 0, spoof);
 		return (FW_DROP);
 	}
 
@@ -2639,7 +2796,7 @@ honeyd_route_packet(struct ip_hdr *ip, u_int iplen,
 		 */
 		tmpl = template_find_best(addr_ntoa(&host), ip, iplen);
 		honeyd_delay_packet(tmpl, ip, iplen, &host, NULL, delay,
-		    DELAY_UNREACH);
+		    DELAY_UNREACH, no_spoof);
 		return (FW_DROP);
 	}
 
@@ -2647,7 +2804,7 @@ honeyd_route_packet(struct ip_hdr *ip, u_int iplen,
 	if (rte != NULL && rte->type == ROUTE_TUNNEL) {
 		honeyd_delay_packet(NULL, ip, iplen,
 		    &rte->tunnel_src, &rte->tunnel_dst,
-		    delay, DELAY_TUNNEL);
+		    delay, DELAY_TUNNEL, no_spoof);
 		return (FW_DROP);
 	}
 
@@ -2694,7 +2851,7 @@ honeyd_input(const struct interface *inter, struct ip_hdr *ip, u_short iplen)
 		}
 		if (tmpl != NULL && tmpl->flags & TEMPLATE_EXTERNAL)
 			flags |= DELAY_ETHERNET;
-		honeyd_delay_packet(NULL, ip, iplen, NULL, NULL, delay, flags);
+		honeyd_delay_packet(NULL, ip, iplen, NULL, NULL, delay, flags, no_spoof);
 		return;
 	}
 
@@ -2775,7 +2932,7 @@ honeyd_input(const struct interface *inter, struct ip_hdr *ip, u_short iplen)
 		flags |= DELAY_ETHERNET;
 
 	/* Delay the packet if necessary, otherwise deliver it directly */
-	honeyd_delay_packet(NULL, ip, iplen, NULL, NULL, delay, flags);
+	honeyd_delay_packet(NULL, ip, iplen, NULL, NULL, delay, flags, no_spoof);
 }
 
 
@@ -2951,6 +3108,10 @@ main(int argc, char *argv[])
 			honeyd_rrdtool_path = optarg;
 			break;
 
+		case 'A':
+			honeyd_webserver_address = optarg;
+			break;
+
 		case 'W':
 			honeyd_webserver_port = atoi(optarg);
 			if (honeyd_webserver_port == 0) {
@@ -3076,6 +3237,10 @@ main(int argc, char *argv[])
 		printf("%s\n", PATH_HONEYDINCLUDE);
 		exit(0);
 	}
+	if (honeyd_show_data_dir) {
+		printf("%s\n", PATH_HONEYDDATA);
+		exit(0);
+	}
 
 	argc -= optind;
 	argv += optind;
@@ -3108,9 +3273,11 @@ main(int argc, char *argv[])
 	router_init();
 	plugins_config_init();
 
-	if (stats_username != NULL)
-		stats_init(&stats_dst, stats_port,
+	if (stats_username != NULL) {
+		stats_init();
+		stats_init_collect(&stats_dst, stats_port,
 		    stats_username, stats_password);
+	}
 
 	personality_init();
 	xprobe_personality_init();
@@ -3147,8 +3314,22 @@ main(int argc, char *argv[])
 	if (want_unittest)
 		unittest();
 
-	if ((honeyd_ip = ip_open()) == NULL)
-		err(1, "ip_open");
+	if ((honeyd_ip = ip_open()) == NULL) {
+		/* 
+		 * We ignore this error if a user just wants to verify
+		 * the configuration - some configs will not load without
+		 * this call succeeding.
+		 */
+		if (!honeyd_verify_config)
+			err(1, "ip_open");
+	}
+
+	if (honeyd_verify_config) {
+		extern int interface_verify_config;
+		
+		/* Make sure that we do not open interfaces for real */
+		interface_verify_config = 1;
+	}
 
 	/* Initialize the specified interfaces */
 	if (ninterfaces == 0)
@@ -3163,9 +3344,11 @@ main(int argc, char *argv[])
 	pyextend_init();
 
 	/* Start our web server */
-	if (honeyd_is_webserver_enabled())
-		pyextend_webserver_init(honeyd_webserver_port,
-		    honeyd_webserver_root);
+	if (!honeyd_verify_config && honeyd_is_webserver_enabled())
+		pyextend_webserver_init(
+			honeyd_webserver_address,
+			honeyd_webserver_port,
+			honeyd_webserver_root);
 #endif
 
 	/* Reads in the ethernet codes and indexes them for use in config */
@@ -3175,6 +3358,10 @@ main(int argc, char *argv[])
 	if (config.config != NULL)
 		config_read(config.config);
 
+	/* Just verify the configuration - exit with success */
+	if (honeyd_verify_config)
+		errx(0, "parsing configuration file successful");
+
 	/* Attach the UI interface */
 	ui_init();
 	
@@ -3201,7 +3388,7 @@ main(int argc, char *argv[])
 	/* Start Honeyd in the background if necessary */
 	if (!honeyd_debug) {
 		setlogmask(LOG_UPTO(LOG_INFO));
-		
+
 		fprintf(stderr, "Honeyd starting as background process\n");
 		if (daemon(1, 0) < 0) {
 			unlink(PIDFILE);
