@@ -2,32 +2,7 @@
  * Copyright (c) 2003 Niels Provos <provos@citi.umich.edu>
  * All rights reserved.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * <LICENSEHERE>
  */
 /*
  * Copyright (c) 2002 Marius Aamodt Eriksen <marius@monkey.org>
@@ -88,6 +63,7 @@
 #include <syslog.h>
 #include <pwd.h>
 #include <dnet.h>
+#include <stdarg.h>
 
 #undef timeout_pending
 #undef timeout_initialized
@@ -98,6 +74,7 @@
 #include "template.h"
 #include "subsystem.h"
 #include "fdpass.h"
+#include "honeyd_overload.h"
 
 #undef DEBUG
 #ifdef DEBUG
@@ -115,6 +92,21 @@
 	memcpy(&(x)->sockaddr, &(y)->sa, (y)->salen); \
 } while (0)
 
+/*
+ * Used to protect a file descriptor if it is used directly to communicate
+ * with.  In the long run, we will have to make new socket pairs for
+ * communication because otherwise we get messed up with multiple processes
+ * using the same file descriptor and then interfere.
+ */
+
+#define PROTECT(x) do { \
+	(x)->flags |= FD_INTERNAL_USE; \
+} while (0)
+
+#define UNPROTECT(x) do { \
+	(x)->flags &= ~FD_INTERNAL_USE; \
+} while (0)
+
 /* Variables figure out where we are */
 
 #define DECLARE(name, ret, args) static ret (*libc_##name) args
@@ -134,6 +126,7 @@ DECLARE(recvfrom, ssize_t, (int, void *, size_t, int, struct sockaddr *,
 DECLARE(sendto, ssize_t, (int, const void *, size_t, int,
 	    const struct sockaddr *, socklen_t));
 DECLARE(sendmsg, ssize_t, (int s, const struct msghdr *msg, int flags));
+DECLARE(recvmsg, ssize_t, (int s, struct msghdr *msg, int flags));
 
 DECLARE(select, int, (int, fd_set *, fd_set *, fd_set *, struct timeval *));
 DECLARE(poll, int, (struct pollfd *, int, int));
@@ -141,6 +134,10 @@ DECLARE(poll, int, (struct pollfd *, int, int));
 DECLARE(accept, int, (int, struct sockaddr *, socklen_t *));
 DECLARE(dup, int, (int));
 DECLARE(dup2, int, (int, int));
+DECLARE(fcntl, int, (int, int, ...));
+#if defined(HAVE_KQUEUE) && 0
+DECLARE(kqueue,int, (void));
+#endif
 
 ssize_t atomicio(ssize_t (*)(), int, void *, size_t);
 
@@ -160,6 +157,9 @@ ssize_t atomicio(ssize_t (*)(), int, void *, size_t);
 #define FD_CONNECTED	0x04
 #define FD_CONNECTING	0x08
 
+#define FD_INTERNAL_USE	0x80
+#define FD_GETSOCKNAME	0x40	/* supports only getsockname */
+
 struct fd {
 	TAILQ_ENTRY(fd) next;
 
@@ -172,13 +172,20 @@ struct fd {
 	int type;
 	int protocol;
 
-	struct sockaddr_storage sa;
+	struct sockaddr_storage sa;	/* address we bound to */
 	socklen_t salen;
 
-	struct sockaddr_storage rsa;
+	struct sockaddr_storage rsa;	/* remote address */
 	socklen_t rsalen;
+
+	struct sockaddr_storage lsa;	/* address we are representing */
+	socklen_t lsalen;
 };
 
+/* Prototypes */
+
+static void free_fd(struct fd *nfd);
+
 #define INIT do { \
 	if (!initalized) \
 		honeyd_init(); \
@@ -186,12 +193,17 @@ struct fd {
 
 static TAILQ_HEAD(fdqueue, fd) fds;
 static int initalized;
+static int magic_fd;
 
 static void
 honeyd_init(void)
 {
 	void *dh;
 
+	magic_fd = atoi(getenv(SUBSYSTEM_MAGICFD));
+	if (magic_fd <= 0)
+		errx(1, "[honeyd_overload] cannot find magic fd");
+
 #ifdef NODLOPEN
 	dh = (void *) -1L;
 #else
@@ -214,15 +226,21 @@ honeyd_init(void)
 
 	GETADDR(sendto);
 	GETADDR(sendmsg);
+	GETADDR(recvmsg);
 
 	GETADDR(select);
 	GETADDR(poll);
 
 	GETADDR(dup);
 	GETADDR(dup2);
+	GETADDR(fcntl);
 
 	GETADDR(accept);
 
+#if defined(HAVE_KQUEUE) && 0
+	GETADDR(kqueue);
+#endif
+
 	/* Do the rest here */
 	TAILQ_INIT(&fds);
 
@@ -301,7 +319,11 @@ clone_fd(struct fd *ofd, int fd)
 
 	nfd->flags = ofd->flags;
 
-	nfd->their_fd = ofd->their_fd;
+	nfd->their_fd = (*libc_dup)(ofd->their_fd);
+	if (nfd->their_fd == -1) {
+		free_fd(nfd);
+		return (NULL);
+	}
 
 	nfd->sa = ofd->sa;
 	nfd->salen = ofd->salen;
@@ -316,14 +338,16 @@ send_cmd(struct subsystem_command *cmd)
 {
 	char res;
 
-	if (atomicio(write, SUBSYSTEM_MAGICFD, cmd,
+	if (atomicio(write, magic_fd, cmd,
 		sizeof(struct subsystem_command)) !=
 	    sizeof(struct subsystem_command)) {
+		DPRINTF((stderr, "%s: write failed\n", __func__));
 		errno = EBADF;
 		return (-1);
 	}
 
-	if (atomicio(read, SUBSYSTEM_MAGICFD, &res, 1) != 1) {
+	if (atomicio(read, magic_fd, &res, 1) != 1) {
+		DPRINTF((stderr, "%s: read failed\n", __func__));
 		errno = EBADF;
 		return (-1);
 	}
@@ -342,14 +366,23 @@ free_fd(struct fd *nfd)
 	free(nfd);
 }
 
+/* Finds an FD as long as the flag_filter does not match */
+
 static struct fd *
-find_fd(int fd)
+find_fd(int fd, int flag_filter)
 {
 	struct fd *nfd;
 
+	/* Never return internal fds */
+	flag_filter |= FD_INTERNAL_USE;
+	
 	TAILQ_FOREACH(nfd, &fds, next)
-	    if (nfd->this_fd == fd)
+	    if (nfd->this_fd == fd) {
+		    /* Do not return the file object if it is protected */
+		    if (nfd->flags & flag_filter)
+			    return (NULL);
 		    return (nfd);
+	    }
 
 	return (NULL);
 }
@@ -398,7 +431,7 @@ listen(int s, int backlog)
 	INIT;
 
 	DPRINTF((stderr, "%s: called on %d\n", __func__, s));
-	if ((nfd = find_fd(s)) == NULL)
+	if ((nfd = find_fd(s, FD_GETSOCKNAME)) == NULL)
 		return ((*libc_listen)(s, backlog));
 
 	if (!(nfd->flags & FD_BOUND)) {
@@ -406,36 +439,44 @@ listen(int s, int backlog)
 		return (-1);
 	}
 
-	SETCMD(&cmd, LISTEN, nfd);
+	SETCMD(&cmd, SUB_LISTEN, nfd);
 	if (send_cmd(&cmd) == -1) {
 		errno = EBADF;
 		return (-1);
 	}
 
 	/* Now send them the fd */
-	send_fd(SUBSYSTEM_MAGICFD, nfd->their_fd, NULL, 0);
-	if (atomicio(read, SUBSYSTEM_MAGICFD, &res, 1) != 1) {
+	send_fd(magic_fd, nfd->their_fd, NULL, 0);
+	if (atomicio(read, magic_fd, &res, 1) != 1) {
 		errno = EBADF;
 		return (-1);
 	}
 
-	close(nfd->their_fd);
+	(*libc_close)(nfd->their_fd);
 	nfd->their_fd = -1;
 
 	return (0);
 }
 
+/*
+ * Protocol:
+ * 1. send bind command
+ * 2. read the allocated port number
+ */
+
 int
 bind(int s, const struct sockaddr *name, socklen_t namelen)
 {
 	struct fd *nfd;
 	struct subsystem_command cmd;
+	u_short port;
 
 	INIT;
 
-	DPRINTF((stderr, "%s: called\n", __func__));
+	DPRINTF((stderr, "%s: called: fd %d familiy %d\n",
+		    __func__, s, name->sa_family));
 
-	if ((nfd = find_fd(s)) == NULL)
+	if ((nfd = find_fd(s, FD_GETSOCKNAME)) == NULL)
 		return ((*libc_bind)(s, name, namelen));
 
 	if (namelen >= sizeof(struct sockaddr_storage)) {
@@ -446,17 +487,46 @@ bind(int s, const struct sockaddr *name, socklen_t namelen)
 	nfd->salen = namelen;
 	memcpy(&nfd->sa, name, namelen);
 
-	SETCMD(&cmd, BIND, nfd);
+	SETCMD(&cmd, SUB_BIND, nfd);
 
 	if (send_cmd(&cmd) == -1) {
 		errno = EADDRINUSE;
 		return (-1);
 	}
 
+	if (atomicio(read, magic_fd, &port, sizeof(port))
+	    != sizeof(port)) {
+		errno = EBADF;
+		return (-1);
+	} else {
+		/* Record local port information */
+		struct sockaddr *sa = (struct sockaddr *)(&nfd->sa);
+		switch (sa->sa_family) {
+		case AF_INET: {
+			struct sockaddr_in *sin = (struct sockaddr_in *)sa;
+			sin->sin_port = htons(port);
+			break;
+		}
+#ifdef AF_INET6
+		case AF_INET6: {
+			struct sockaddr_in6 *sin = (struct sockaddr_in6 *)sa;
+			sin->sin6_port = htons(port);
+			break;
+		}
+#endif
+		default:
+			DPRINTF((stderr,
+				    "%s: bad socket family on %d: %d\n",
+				    __func__, s, sa->sa_family));
+			break;
+		}
+	}
+
 	nfd->flags &= ~FD_UNBOUND;
 	nfd->flags = FD_BOUND;
 
-	DPRINTF((stderr, "%s: socket %d bound\n", __func__, s));
+	DPRINTF((stderr, "%s: socket %d bound at port %d\n",
+		    __func__, s, port));
 
 	return (0);
 }
@@ -470,12 +540,12 @@ close(int fd)
 	INIT;
 
 	/* Don't close the magic file descriptor that points back to us */
-	if (fd == SUBSYSTEM_MAGICFD) {
+	if (fd == magic_fd) {
 		errno = EBADF;
 		return (-1);
 	}
 
-	if ((nfd = find_fd(fd)) == NULL)
+	if ((nfd = find_fd(fd, 0)) == NULL)
 		return ((*libc_close)(fd));
 
 	DPRINTF((stderr, "%s: with %d, flags %x\n", __func__,
@@ -484,7 +554,7 @@ close(int fd)
 
 	/* XXX - need to tell honeyd about close in other cases */
 	if (nfd->flags & FD_BOUND) {
-		SETCMD(&cmd, CLOSE, nfd);
+		SETCMD(&cmd, SUB_CLOSE, nfd);
 		send_cmd(&cmd);
 	}
 
@@ -504,9 +574,10 @@ connect(int s, const struct sockaddr *name, socklen_t namelen)
 
 	INIT;
 
-	DPRINTF((stderr, "%s: called\n", __func__));
+	DPRINTF((stderr, "%s: called: %d: %p %d\n",
+		    __func__, s, name, namelen));
 
-	if ((nfd = find_fd(s)) == NULL)
+	if ((nfd = find_fd(s, FD_GETSOCKNAME)) == NULL)
 		return ((*libc_connect)(s, name, namelen));
 
 	/* Report an error if the socket is connected already */
@@ -535,26 +606,26 @@ connect(int s, const struct sockaddr *name, socklen_t namelen)
 		return (-1);
 	}
 
-	SETCMD(&cmd, CONNECT, nfd);
+	SETCMD(&cmd, SUB_CONNECT, nfd);
 	cmd.rlen = namelen;
 	memcpy(&cmd.rsockaddr, name, namelen);
 	/* Copy local address, too */
 	cmd.len = nfd->salen;
 	memcpy(&cmd.sockaddr, &nfd->sa, nfd->salen);
 	if (send_cmd(&cmd) == -1) {
-		close(pair[0]);
-		close(pair[1]);
+		(*libc_close)(pair[0]);
+		(*libc_close)(pair[1]);
 		errno = ENETUNREACH;
 		return (-1);
 	}
 
 	/* Send special communication fd */
-	send_fd(SUBSYSTEM_MAGICFD, pair[1], NULL, 0);
-	close(pair[1]);
+	send_fd(magic_fd, pair[1], NULL, 0);
+	(*libc_close)(pair[1]);
 
 	if (atomicio(read, pair[0], &res, sizeof(res)) != sizeof(res)){
-		close(pair[0]);
-		close(pair[1]);
+		(*libc_close)(pair[0]);
+		(*libc_close)(pair[1]);
 		DPRINTF((stderr, "%s: failure to send fd\n", __func__));
 		errno = EBADF;
 		return (-1);
@@ -567,7 +638,7 @@ connect(int s, const struct sockaddr *name, socklen_t namelen)
 
 #if 0
 	/* This is the point where we need to check non-blocking IO */
-	flags = fcntl(nfd->this_fd, F_GETFL, NULL);
+	flags = (*libc_fcntl)(nfd->this_fd, F_GETFL, NULL);
 	if (flags != -1 && (flags & O_NONBLOCK)) {
 		fcntl(pair[0], F_SETFL, O_NONBLOCK);
 	}
@@ -580,10 +651,10 @@ connect(int s, const struct sockaddr *name, socklen_t namelen)
 	}
 
  	/* Now we can close the special communication fds */
-	close(pair[0]);
-	close(pair[1]);
+	(*libc_close)(pair[0]);
+	(*libc_close)(pair[1]);
 
-	close(nfd->their_fd);
+	(*libc_close)(nfd->their_fd);
 	nfd->their_fd = -1;
 
 	nfd->salen = sizeof(si);
@@ -653,7 +724,7 @@ recvfrom(int sock, void *buf, size_t len, int flags, struct sockaddr *from,
 
 	DPRINTF((stderr, "%s: read %d\n", __func__, ret));
 
-	if (from != NULL && (nfd = find_fd(sock)) != NULL) {
+	if (from != NULL && (nfd = find_fd(sock, FD_GETSOCKNAME)) != NULL) {
 		if (flen < nfd->rsalen)
 			goto out;
 		memcpy(from, &nfd->rsa, nfd->rsalen);
@@ -673,7 +744,7 @@ sendto(int sock, const void *buf, size_t len, int flags,
 
 	INIT;
 
-	nfd = find_fd(sock);
+	nfd = find_fd(sock, FD_GETSOCKNAME);
 
 	if (nfd == NULL)
 		return ((*libc_sendto)(sock, buf, len, flags, to, tolen));
@@ -691,7 +762,8 @@ sendto(int sock, const void *buf, size_t len, int flags,
 
 	ret = (*libc_sendto)(sock, buf, len, flags, NULL, 0);
 
-	DPRINTF((stderr, "%s: called: %d: %p,%d -> %d (%s)\n", __func__,
+	DPRINTF((stderr, "%s: called: %d: len %d: sa %p,%d -> %d (%s)\n",
+		    __func__, len,
 		    sock, to, tolen,
 		    ret, ret != -1 ? "no error" : strerror(errno)));
 
@@ -702,37 +774,110 @@ int
 getsockname(int sock, struct sockaddr *to, socklen_t *tolen)
 {
 	struct fd *nfd;
-	socklen_t copylen;
+	struct sockaddr *src;
+	socklen_t srclen;
+	
 
 	INIT;
 
-	nfd = find_fd(sock);
-
+	nfd = find_fd(sock, 0);
 	if (nfd == NULL)
 		return ((*libc_getsockname)(sock, to, tolen));
 
 	DPRINTF((stderr, "%s: called: %d: %p,%d\n", __func__,
 		    sock, to, *tolen));
 
-	copylen = nfd->salen;
-	if (*tolen < copylen)
-		copylen = *tolen;
+	/*
+	 * Get the real local address if possible, otherwise return
+	 * the address we bound to.
+	 */
+	if (nfd->lsalen) {
+		src = (struct sockaddr *)&nfd->lsa;
+		srclen = nfd->lsalen;
+	} else {
+		src = (struct sockaddr *)&nfd->sa;
+		srclen = nfd->salen;
+	}
+	
+	if (*tolen < srclen)
+		srclen = *tolen;
 	else
-		*tolen = copylen;
-	memcpy(to, &nfd->sa, copylen);
+		*tolen = srclen;
+	memcpy(to, src, srclen);
 
 	return (0);
 }
 
+ssize_t
+recvmsg(int sock, struct msghdr *msg, int flags)
+{
+	struct fd *nfd;
+	ssize_t ret = -1;
+	size_t len, off;
+	int i;
+	void *data;
+
+	INIT;
+
+	nfd = find_fd(sock, FD_GETSOCKNAME);
+
+	if (nfd == NULL)
+		return ((*libc_recvmsg)(sock, msg, flags));
+
+	errno = EINVAL;
+	DPRINTF((stderr, "%s: called: %d: %p, %d\n", __func__, sock,
+		    msg, flags));
+
+	/* We do not currently support these flags */
+	if ( flags & (MSG_OOB|MSG_PEEK) ) {
+		errno = EINVAL;
+		return (-1);
+	}
+
+	/* We would like to know how much data we can read. */
+	len = 0;
+	for ( i = 0; i < msg->msg_iovlen; i++ ) {
+		len += msg->msg_iov[i].iov_len;
+	}
+	if ((data = malloc(len)) == NULL) {
+		errno = ENOBUFS;
+		return (-1);
+	}
+
+	/* Now we have successfully converted the call into a recvmsg call */
+	ret = recvfrom(sock, data, len, flags,
+	    msg->msg_name, &msg->msg_namelen);
+
+	if (ret == -1)
+		goto out;
+
+	/* Copy the data back into the provided memory buffers */
+	for ( i = 0, off = 0; i < msg->msg_iovlen && off < ret; i++ ) {
+		ssize_t avail = msg->msg_iov[i].iov_len;
+		if (avail > ret - off)
+			avail = ret - off;
+		memcpy(msg->msg_iov[i].iov_base, data + off, avail);
+		off += avail;
+	}
+
+ out:
+	free(data);
+
+	return (ret);
+}
+
 ssize_t
 sendmsg(int sock, const struct msghdr *msg, int flags)
 {
 	struct fd *nfd;
 	ssize_t ret = -1;
+	size_t len, off;
+	int i;
+	void *data;
 
 	INIT;
 
-	nfd = find_fd(sock);
+	nfd = find_fd(sock, FD_GETSOCKNAME);
 
 	if (nfd == NULL)
 		return ((*libc_sendmsg)(sock, msg, flags));
@@ -741,6 +886,35 @@ sendmsg(int sock, const struct msghdr *msg, int flags)
 	DPRINTF((stderr, "%s: called: %d: %p, %d\n", __func__, sock,
 		    msg, flags));
 
+	/* We do not currently support these flags */
+	if ( flags & (MSG_OOB|MSG_DONTROUTE) ) {
+		errno = EINVAL;
+		return (-1);
+	}
+
+	/* We just gather the data and then send it as bulk */
+	len = 0;
+	for ( i = 0; i < msg->msg_iovlen; i++ ) {
+		len += msg->msg_iov[i].iov_len;
+	}
+	if ((data = malloc(len)) == NULL) {
+		errno = ENOBUFS;
+		return (-1);
+	}
+
+	/* Copy all the data into our single buffer */
+	for ( i = 0, off = 0; i < msg->msg_iovlen; i++ ) {
+		memcpy(data + off,
+		    msg->msg_iov[i].iov_base,
+		    msg->msg_iov[i].iov_len);
+		off += msg->msg_iov[i].iov_len;
+	}
+
+	/* Now we have successfully converted the call into a sendmsg call */
+	ret = sendto(sock, data, len, 0, msg->msg_name, msg->msg_namelen);
+
+	free(data);
+
 	return (ret);
 }
 
@@ -754,6 +928,112 @@ setsockopt(int sock, int level, int optname, const void *optval,
 	return ((*libc_setsockopt)(sock, level, optname, optval, option));
 }
 
+int
+fcntl(int fd, int cmd, ...)
+{
+	struct fd *nfd;
+	int argument, i;
+	int req_special = 0;
+	int ret = -1;
+	va_list ap;
+
+	va_start(ap, cmd);
+
+	INIT;
+
+	/* Some fcntl commands require our special attention */
+	if (cmd == F_DUPFD || cmd == F_SETFD || cmd == F_XXX_GETSOCK)
+		req_special = 1;
+
+	if (!req_special || (nfd = find_fd(fd, FD_GETSOCKNAME)) == NULL) {
+		struct flock *flock;
+
+		switch (cmd) {
+		case F_GETLK:
+		case F_SETLK:
+		case F_SETLKW:
+			flock = va_arg(ap, struct flock *);
+			return (*libc_fcntl)(fd, cmd, flock);
+		default:
+			argument = va_arg(ap, int);
+			return (*libc_fcntl)(fd, cmd, argument);
+		}
+
+	}
+
+	/*
+	 * A special hook for a subsystem to get the local address
+	 * information for a connected socket.
+	 */
+	if (cmd == F_XXX_GETSOCK) {
+		struct sockaddr *sa = va_arg(ap, struct sockaddr *);
+		socklen_t *psalen = va_arg(ap, socklen_t *);
+		va_end(ap);
+
+		DPRINTF((stderr, "%s: called: %d XXX_GETSOCK\n",
+			    __func__, fd));
+
+		if (nfd->lsalen == 0) {
+			errno = EBADF;
+			return (-1);
+		}
+		
+		if (*psalen < nfd->lsalen) {
+			errno = EINVAL;
+			return (-1);
+		}
+
+		*psalen = nfd->lsalen;
+		memcpy(sa, &nfd->lsa, nfd->lsalen);
+
+		return (0);
+	}
+
+	/* Get the fd argument from fcntl */
+	argument = va_arg(ap, int);
+
+	va_end(ap);
+
+	switch (cmd) {
+	case F_DUPFD: {
+		DPRINTF((stderr, "%s: called: %d dup > %d\n",
+			    __func__, fd, argument));
+
+		/* 
+		 * Try to find an unused descriptor that is higher
+		 * than the desired.  XXX: Remove hard coded limit;
+		 */
+		for ( i = argument; i < 4096; i++ ) {
+			/* fcntl will fail on unallocated file descriptor */
+			if ((*libc_fcntl)(i, F_GETFD) == -1)
+				break;
+		}
+
+		/* Out of file descriptors */
+		if ( i == 4096 ) {
+			errno = EMFILE;
+			return (-1);
+		}
+
+		/* Duplicate the file descriptor as desired */
+		return (dup2(fd, i));
+	}
+	case F_SETFD: {
+		DPRINTF((stderr, "%s: called: %d setfd: %d\n",
+			    __func__, fd, argument));
+		ret = (*libc_fcntl)(fd, F_SETFD, argument);
+		if (ret != -1 && nfd->their_fd != -1)
+			ret = (*libc_fcntl)(nfd->their_fd, F_SETFD, argument);
+		break;
+	}
+	default:
+		DPRINTF((stderr, "%s: unknown fcntl command: %d\n",
+			    __func__, cmd));
+	}
+
+	return ret;
+}
+
 int
 dup(int oldfd)
 {
@@ -772,8 +1052,8 @@ dup(int oldfd)
 	if (newfd == -1)
 		return (-1);
 
-	nfd = find_fd(oldfd);
-	if (clone_fd(nfd, newfd) == NULL) {
+	nfd = find_fd(oldfd, 0);
+	if (nfd != NULL && clone_fd(nfd, newfd) == NULL) {
 		(*libc_close)(newfd);
 		errno = EMFILE;
 		return (-1);
@@ -793,7 +1073,7 @@ dup2(int oldfd, int newfd)
 	DPRINTF((stderr, "%s: called: %d -> %d\n", __func__, oldfd, newfd));
 
 	/* Prevent overwriting of our control fd */
-	if (newfd == SUBSYSTEM_MAGICFD) {
+	if (newfd == magic_fd) {
 		errno = EBADF;
 		return (-1);
 	}
@@ -805,8 +1085,8 @@ dup2(int oldfd, int newfd)
 	if (newfd == -1)
 		return (-1);
 
-	nfd = find_fd(oldfd);
-	if (clone_fd(nfd, newfd) == NULL) {
+	nfd = find_fd(oldfd, 0);
+	if (nfd != NULL && clone_fd(nfd, newfd) == NULL) {
 		(*libc_close)(newfd);
 		errno = EMFILE;
 		return (-1);
@@ -819,22 +1099,26 @@ int
 accept(int sock, struct sockaddr *addr, socklen_t *addrlen)
 {
 	struct fd *nfd;
-	struct sockaddr_storage sa;
+	struct bundle bundle;
 	socklen_t salen;
 	int fd;
 
 	INIT;
 
-	nfd = find_fd(sock);
+	nfd = find_fd(sock, FD_GETSOCKNAME);
 
-	DPRINTF((stderr, "%s: called: %p\n", __func__, nfd));
+	DPRINTF((stderr, "%s: called: %d -> %p\n", __func__, sock, nfd));
 
 	if (nfd == NULL)
 		return (*libc_accept)(sock, addr, addrlen);
 
 	/* Get a connection from Honeyd */
-	salen = sizeof(sa);
-	if ((fd = receive_fd(sock, &sa, &salen)) == -1) {
+	salen = sizeof(bundle);
+	/* Do not intercept calls on this */
+	PROTECT(nfd);
+	fd = receive_fd(sock, &bundle, &salen);
+	UNPROTECT(nfd);
+	if (fd == -1) {
 		DPRINTF((stderr, "%s: failed\n", __func__));
 		return (-1);
 	}
@@ -842,14 +1126,34 @@ accept(int sock, struct sockaddr *addr, socklen_t *addrlen)
 	/* XXX - something good happened! */
 	DPRINTF((stderr, "%s: got %d (salen %d)\n", __func__, fd, salen));
 
-	/* Store for later */
-	nfd->rsalen = salen;
-	memcpy(&nfd->rsa, &sa, salen);
-
 	if (addr != NULL) {
-		*addrlen = salen;
-		memcpy(addr, &sa, salen);
+		*addrlen = sizeof(bundle.src);
+		memcpy(addr, &bundle.src, sizeof(bundle.src));
 	}
 
+	/* create a new mapping fd for the accepted connection */
+	nfd = new_fd(fd);
+	nfd->flags |= FD_GETSOCKNAME;
+
+	/* Store for later */
+	nfd->rsalen = sizeof(bundle.src);
+	memcpy(&nfd->rsa, &bundle.src, nfd->rsalen);
+
+	nfd->lsalen = sizeof(bundle.dst);
+	memcpy(&nfd->lsa, &bundle.dst, nfd->lsalen);
+
 	return (fd);
 }
+
+#if 0
+
+/* We DO NOT support kqueue */
+
+int
+kqueue(void)
+{
+	errno = EOPNOTSUPP;
+	return (-1);
+}
+
+#endif
