@@ -70,7 +70,7 @@ void pf_osfp_initialize(void);
 int pf_osfp_match(struct pf_osfp_enlist *, pf_osfp_t);
 struct pf_osfp_enlist *pf_osfp_fingerprint_hdr(const struct ip_hdr *, const struct tcp_hdr *);
 
-void honeyd_osfp_input(u_char *, u_int, void *);
+void honeyd_osfp_input(struct tuple *, u_char *, u_int, void *);
 static struct osfp *honeyd_osfp_cache(const struct ip_hdr *);
 
 SPLAY_HEAD(osfptree, osfp) osfp_buckets[OSFP_HASHSIZE];
@@ -80,7 +80,7 @@ osfp_compare(struct osfp *a, struct osfp *b)
 {
 	if (a->src < b->src)
 		return (-1);
-	if (a->src > b->src)
+	else if (a->src > b->src)
 		return (1);
 	return (0);
 }
@@ -157,7 +157,7 @@ honeyd_osfp_cache_insert(const struct ip_hdr *ip, struct pf_osfp_enlist *list)
 		}
 
 		entry->src = ip->ip_src;
-		timeout_set(&entry->timeout, honeyd_osfp_timeout, entry);
+		evtimer_set(&entry->timeout, honeyd_osfp_timeout, entry);
 		root = honeyd_osfp_hash(ip);
 
 		SPLAY_INSERT(osfptree, root, entry);
@@ -167,7 +167,7 @@ honeyd_osfp_cache_insert(const struct ip_hdr *ip, struct pf_osfp_enlist *list)
 	
 	timerclear(&tv);
 	tv.tv_sec = OSFP_TIMEOUT;
-	timeout_add(&entry->timeout, &tv);
+	evtimer_add(&entry->timeout, &tv);
 }
 
 static struct osfp *
@@ -190,7 +190,7 @@ honeyd_osfp_cache(const struct ip_hdr *ip)
 	/* Update timeout */
 	timerclear(&tv);
 	tv.tv_sec = OSFP_TIMEOUT;
-	timeout_add(&entry->timeout, &tv);
+	evtimer_add(&entry->timeout, &tv);
 
 	return (entry);
 }
@@ -214,7 +214,7 @@ honeyd_osfp_match(const struct ip_hdr *ip, pf_osfp_t fp)
 }
 
 void
-honeyd_osfp_input(u_char *pkt, u_int plen, void *arg)
+honeyd_osfp_input(struct tuple *conhdr, u_char *pkt, u_int plen, void *arg)
 {
 	const struct ip_hdr *ip = (struct ip_hdr *)pkt;
 	const struct tcp_hdr *tcp;
