@@ -66,14 +66,30 @@
 
 /* ET - Moved SPLAY_HEAD to personality.h so xprobe_assoc.c could use it. */
 int npersons;
+
 /* ET - global from honeyd.c */
 struct personate person_drop = {};
+static struct event personality_time_ev;
+static struct timeval tv_periodic;
 
 SPLAY_GENERATE(perstree, personality, node, perscompare);
 
 /* ET - For the Xprobe fingerprint tree */
 SPLAY_GENERATE(xp_fprint_tree, xp_fingerprint, node, xp_fprint_compare);
 
+static void
+personality_time_evcb(int fd, short what, void *arg)
+{
+	struct event *ev = arg;
+	struct timeval tv;
+
+	gettimeofday(&tv_periodic, NULL);
+
+	timerclear(&tv);
+	tv.tv_usec = 100000;	/* every 100 ms */
+	evtimer_add(ev, &tv);
+}
+
 void
 xprobe_personality_init(void)
 {
@@ -85,6 +101,11 @@ personality_init(void)
 {
 	npersons = 0;
 	SPLAY_INIT(&personalities);
+
+	/* Start a timer that keeps track of the current system time */
+	evtimer_set(&personality_time_ev,
+	    personality_time_evcb, &personality_time_ev);
+	personality_time_evcb(-1, EV_TIMEOUT, &personality_time_ev);
 }
 
 struct personality *
@@ -376,38 +397,42 @@ get_next_isn(struct template *tmpl, const struct personality *person)
         return (prev_isn + ((((num_isns - 1) % 5) + 1) * (uint32_t)a - 1)*gcd);
 }
 
-#define TIME_CORRECT(x) do { \
-	tmp.tv_sec = 0; \
-	tmp.tv_usec = tmp.tv_usec % (x); \
-	timersub(&tmpl->tv, &tmp, &tmpl->tv); \
+#define TIME_CORRECT(x,y) do { \
+	(y)->tv_sec = 0; \
+	(y)->tv_usec = (y)->tv_usec % (x); \
+	timersub(&tmpl->tv, y, &tmpl->tv); \
 } while (0)
 
+/* Get the correct time for this personality */
 
-uint32_t
-tcp_personality_seq(struct template *tmpl, struct personality *person)
+void
+personality_time(struct template *tmpl, struct timeval *diff)
+{
+	uint32_t ms, old_ms;
+
+	timersub(&tv_periodic, &tmpl->tv_real, diff);
+	tmpl->tv_real = tv_periodic;
+
+	old_ms = ms = diff->tv_sec * 10000 + (diff->tv_usec / 100);
+	ms *= tmpl->drift;
+
+	diff->tv_sec = ms / 10000;
+	diff->tv_usec = (ms % 10000) * 100;
+
+	timeradd(&tmpl->tv, diff, &tmpl->tv);
+}
+
+int
+tcp_personality_time(struct template *tmpl, struct timeval *diff)
 {
-	struct timeval tv, tmp;
 	extern rand_t *honeyd_rand;
+	struct personality *person = tmpl->person;
 	int slowhz;
 
-	tmpl->seqcalls++;
+	if (person == NULL)
+		return (-1);
 
-	if (!timerisset(&tmpl->tv)) {
-		gettimeofday(&tmpl->tv, NULL);
-		if (tmpl->timestamp == 0)
-			tmpl->timestamp = rand_uint32(honeyd_rand) % 1728000;
-		if (tmpl->seq == 0) {
-			if (person->seqt == SEQ_CONSTANT && person->valset)
-				tmpl->seq = person->val;
-			else
-				tmpl->seq = rand_uint32(honeyd_rand);
-		}
-		return (tmpl->seq);
-	}
-
-	gettimeofday(&tv, NULL);
-	timersub(&tv, &tmpl->tv, &tmp);
-	tmpl->tv = tv;
+	personality_time(tmpl, diff);
 
 	if (person->tstamphz) {
 		int tstamphz = person->tstamphz;
@@ -421,8 +446,8 @@ tcp_personality_seq(struct template *tmpl, struct personality *person)
 		 * counted next time.
 		 */
 		ticks = 1000000L / tstamphz;
-		slowhz = tmp.tv_sec * tstamphz + tmp.tv_usec / ticks;
-		TIME_CORRECT(ticks);
+		slowhz = diff->tv_sec * tstamphz + diff->tv_usec / ticks;
+		TIME_CORRECT(ticks, diff);
 
 		tmpl->timestamp += slowhz;
 	} else {
@@ -440,25 +465,72 @@ tcp_personality_seq(struct template *tmpl, struct personality *person)
 	 * generated so far.
 	 */
 	switch(person->seqt) {
-	case SEQ_CONSTANT:
-		return (tmpl->seq);
 	case SEQ_TRIVIALTIME:
 	case SEQ_RI:
 		/* No time component.  May be required for high latency */
-		if (tmp.tv_sec > 2) {
+		if (diff->tv_sec > 2) {
 			uint32_t adjust;
 
 			adjust = rand_uint32(honeyd_rand) % 2048;
 			adjust *= (slowhz - 2) * person->gcd;
 			tmpl->seq += adjust;
 		}
+		break;
+	case SEQ_CLASS64K:
+		tmpl->seq += slowhz * 64000;
+		break;
+	case SEQ_I800:
+		tmpl->seq += slowhz * 800;
+		break;
+	default:
+		break;
+	}
+
+	return (slowhz);
+}
+
+uint32_t
+tcp_personality_seq(struct template *tmpl, struct personality *person)
+{
+	struct timeval tmp;
+	extern rand_t *honeyd_rand;
+	int slowhz;
+
+	tmpl->seqcalls++;
+
+	if (!timerisset(&tmpl->tv)) {
+		gettimeofday(&tv_periodic, NULL);
+		tmpl->tv_real = tmpl->tv = tv_periodic;
+		if (tmpl->timestamp == 0)
+			tmpl->timestamp = rand_uint32(honeyd_rand) % 1728000;
+		if (tmpl->seq == 0) {
+			if (person->seqt == SEQ_CONSTANT && person->valset)
+				tmpl->seq = person->val;
+			else
+				tmpl->seq = rand_uint32(honeyd_rand);
+		}
+		return (tmpl->seq);
+	}
+
+	slowhz = tcp_personality_time(tmpl, &tmp);
+
+	/* 
+	 * This is where new ISNs are generated.  The latest ISN is
+	 * stored in tmpl->seq.  tmpl->seqcalls is the number of ISNs
+	 * generated so far.
+	 */
+	switch(person->seqt) {
+	case SEQ_CONSTANT:
+		return (tmpl->seq);
+	case SEQ_TRIVIALTIME:
+	case SEQ_RI:
 		tmpl->seq = get_next_isn(tmpl, person);
 		return (tmpl->seq);
 	case SEQ_CLASS64K:
-		tmpl->seq += (slowhz + 1) * 64000;
+		tmpl->seq += 64000;
 		return (tmpl->seq);
 	case SEQ_I800:
-		tmpl->seq += (slowhz + 1) * 800;
+		tmpl->seq += 800;
 		return (tmpl->seq);
 	case SEQ_RANDOM:
 	default:
@@ -513,6 +585,7 @@ tcp_personality(struct tcp_con *con, uint8_t *pflags, int *pwindow, int *pdf,
 		/* Set the sequence number only on SYN segments */
 		if (con->snd_una == 0 && (flags & TH_SYN))
 			con->snd_una = tcp_personality_seq(tmpl, person);
+
 		/* If we support timestamps, always set them */
 		if (person->tstamphz >= 0 && poptions != NULL)
 			*poptions = default_opts;
@@ -587,10 +660,13 @@ tcp_personality_options(struct tcp_con *con, struct tcp_hdr *tcp,
 			SET(p, &opt, TCP_OPT_WSCALE, 3);
 			break;
 		case 't':
-			if (tmpl != NULL)
+			if (tmpl != NULL) {
+				struct timeval tv;
+				tcp_personality_time(tmpl, &tv);
 				timestamp = htonl(tmpl->timestamp);
-			else
+			} else {
 				timestamp = rand_uint32(honeyd_rand);
+			}
 			opt.opt_data.timestamp[0] = timestamp;
 			opt.opt_data.timestamp[1] = 0;
 			if (con->sawtimestamp)
@@ -916,23 +992,31 @@ parse_tseq(struct personality *pers, int off, char *line)
 		}
 		/* Improve IPID sequencing capability */
 		else if (strncasecmp(p2, "IPID=", 5) == 0) {
-			p2 = strsep(&end, "|");
-			p2 += 5;
-		  
-			if (strcasecmp(p2, "I") == 0)
-				pers->idt = ID_SEQUENTIAL;
-			else if (strcasecmp(p2, "BI") == 0)
-				pers->idt = ID_SEQUENTIAL_BROKEN;
-			else if (strcasecmp(p2, "Z") == 0)
-				pers->idt = ID_ZERO;
-			else if (strcasecmp(p2, "RD") == 0)
-				pers->idt = ID_RANDOM;
-			else if (strcasecmp(p2, "C") == 0)
-				pers->idt = ID_CONSTANT;
-			else if (strcasecmp(p2, "RPI") == 0)
-				pers->idt = ID_RPI;
-			else
-				return -1;
+			int done = 0;
+
+			end += 5;
+			while (!done) {
+				done = 1;
+				p2 = strsep(&end, "|");
+				if (strcasecmp(p2, "I") == 0) {
+					pers->idt = ID_SEQUENTIAL;
+				} else if (strcasecmp(p2, "BI") == 0) {
+					pers->idt = ID_SEQUENTIAL_BROKEN;
+				} else if (strcasecmp(p2, "Z") == 0) {
+					if (end != NULL) {
+						done = 0;
+						continue;
+					}
+					pers->idt = ID_ZERO;
+				} else if (strcasecmp(p2, "RD") == 0) {
+					pers->idt = ID_RANDOM; 
+				} else if (strcasecmp(p2, "C") == 0) {
+					pers->idt = ID_CONSTANT;
+				} else if (strcasecmp(p2, "RPI") == 0) {
+					pers->idt = ID_RPI;
+				} else
+					return -1;
+			}
 
 		} else if (strncasecmp(p2, "TS=", 3) == 0) {
 			/* TCP timestamp sequencing capability */
