@@ -137,7 +137,12 @@ struct personality {
 	uint32_t seqindex_max;
 	/* ET - improve TCP ISN calculation */
 	uint32_t gcd;
+
 	uint32_t val;		/* Used for constant ISNs */
+
+	uint8_t valset:1,
+	        unused:7;
+
 	double seqindex_amin;	/* Values calculated from seqindex_{min,max} */
 	double seqindex_amax;
 	double seqindex_aconst;
@@ -163,6 +168,7 @@ void ip_personality(struct template *, uint16_t *);
 int tcp_personality(struct tcp_con *, uint8_t *, int *, int *,
     uint16_t *, char **);
 void tcp_personality_options(struct tcp_con *, struct tcp_hdr *, char *);
+int tcp_personality_match(struct tcp_con *, int);
 
 int icmp_error_personality(struct template *, struct addr *,
     struct ip_hdr *ip, uint8_t *, uint8_t *, int *, uint8_t *);
