@@ -2,32 +2,7 @@
  * Copyright (c) 2003 Niels Provos <provos@citi.umich.edu>
  * All rights reserved.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * <LICENSEHERE>
  */
 
 #include <sys/types.h>
@@ -82,6 +57,12 @@
 int make_socket(int (*f)(int, const struct sockaddr *, socklen_t), int type,
     char *, uint16_t);
 
+/*
+ * Can be used externally to share Python dictionaries with the user interface
+ */
+PyObject *pyextend_dict_global;
+PyObject *pyextend_dict_local;
+
 /* 
  * Functions that we need to call for this script.
  * This is stateless and shared among connections.
@@ -543,9 +524,14 @@ static PyObject*
 pyextend_log(PyObject *self, PyObject *args)
 {
 	extern FILE *honeyd_servicefp;
-	struct tuple *hdr = current_state->con;
+	struct tuple *hdr;
 	char *string;
 
+	if (current_state == NULL)
+		return (Py_BuildValue("i", -1));
+	
+	hdr = current_state->con;
+	
 	if(!PyArg_ParseTuple(args, "s:read_selector", &string))
 		return (NULL);
 
@@ -804,6 +790,7 @@ pyextend_init(void)
 	pModule = Py_InitModule("honeyd", HoneydMethods);
 	PyModule_AddIntConstant(pModule, "EVENT_ON", 1);
 	PyModule_AddIntConstant(pModule, "EVENT_OFF", 0);
+	PyModule_AddStringConstant(pModule, "version", VERSION);
 }
 
 /* Cleans up all Python stuff when we exit */
@@ -814,6 +801,94 @@ pyextend_exit(void)
 	Py_Finalize();
 }
 
+void
+pyextend_run(struct evbuffer *output, char *command)
+{
+	PyObject *res = NULL, *compiled_code;
+	char *data;
+	int datlen;
+
+	char *preamble = "import StringIO\n"
+	    "import sys\n"
+	    "myout = StringIO.StringIO()\n"
+	    "myerr = StringIO.StringIO()\n"
+	    "saveout = sys.stdout\n"
+	    "saveerr = sys.stderr\n"
+	    "try:\n"
+	    "  sys.stdout = myout\n"
+	    "  sys.stderr = myerr\n"
+	    "  try:\n"
+	    "    %s\n"
+	    "  except:\n"
+	    "    import traceback\n"
+	    "    traceback.print_exc()\n"
+	    "finally:\n"
+	    "  sys.stdout = saveerr\n"
+	    "  sys.stderr = saveerr\n"
+	    "output = \"%%s%%s\" %% (myout.getvalue(), myerr.getvalue())";
+
+	char *code = NULL;
+
+	if (asprintf(&code, preamble, command) == -1)
+		err(1, "%s: asprintf", __func__);
+
+	compiled_code = Py_CompileStringFlags(code, "<filter>",
+	    Py_file_input, 0);
+
+	free(code);
+	
+	if (compiled_code == NULL) {
+		const char *err = "Compilation of Python code failed.\n";
+		evbuffer_add(output, (char *)err, strlen(err));
+		PyErr_Print();
+		return;
+	}
+
+	if (pyextend_dict_local == NULL) {
+		pyextend_dict_local = PyDict_New();
+		assert(pyextend_dict_local != NULL);
+	}
+	
+	if (pyextend_dict_global == NULL) {
+		PyObject *m;
+
+		/* Extract the global dictionary object */
+		if ((m = PyImport_AddModule("__main__")) == NULL) {
+			PyErr_Print();
+			return;
+		}
+
+		if ((pyextend_dict_global = PyModule_GetDict(m)) == NULL) {
+			PyErr_Print();
+			return;
+		}
+		Py_INCREF(pyextend_dict_global);
+
+		if (PyDict_GetItemString(pyextend_dict_global, "__builtins__") == NULL&&
+		    PyDict_SetItemString(pyextend_dict_global, "__builtins__", PyEval_GetBuiltins()) == 0) {
+			Py_DECREF(pyextend_dict_global);
+			pyextend_dict_global = NULL;
+			return;
+		}
+	}
+
+	res = PyEval_EvalCode((PyCodeObject *)compiled_code,
+	    pyextend_dict_global, pyextend_dict_local);
+	Py_DECREF(compiled_code);
+
+	if (res == NULL) {
+		PyErr_Print();
+		return;
+	}
+	Py_DECREF(res);
+
+	res = PyDict_GetItemString(pyextend_dict_local, "output");
+	assert(res != NULL);
+
+	if (PyString_AsStringAndSize(res, &data, &datlen) == 0) 
+		evbuffer_add(output, data, datlen);
+}
+
 #define CHECK_FUNC(f, x) do { \
 	f = PyDict_GetItemString(pDict, x); \
 	if ((f) == NULL || !PyCallable_Check(f)) { \
@@ -1020,8 +1095,8 @@ pyextend_connection_end(struct pystate *state)
  * via a web browser.
  */
 
-static PyObject *pWebServer;
-static PyObject *pFuncRequest;
+static PyObject *pWebServer;	/* web server instance */
+static PyObject *pFuncRequest;	/* handle request function */
 
 static int pyserver_fd = -1;
 static struct event ev_accept;
@@ -1276,7 +1351,7 @@ pyextend_webserver_verify_setup(const char *root_dir)
  */
 
 void
-pyextend_webserver_init(int port, char *root_dir)
+pyextend_webserver_init(char *address, int port, char *root_dir)
 {
 	PyObject *pArgs, *pName, *pModule, *pDict, *pFuncWebInit;
 	char *script = "server";
@@ -1307,7 +1382,7 @@ pyextend_webserver_init(int port, char *root_dir)
 		errx(1, "%s: make_server function returned error.", __func__);
 	}
 
-	pyserver_fd = make_socket(bind, SOCK_STREAM, "0.0.0.0", port);
+	pyserver_fd = make_socket(bind, SOCK_STREAM, address, port);
 
 	if (pyserver_fd == -1) {
 		fprintf(stderr,
@@ -1320,7 +1395,7 @@ pyextend_webserver_init(int port, char *root_dir)
 	if (listen(pyserver_fd, 10) == -1)
 		err(1, "%s: listen", __func__);
 
-	syslog(LOG_NOTICE, "HTTP server listening on port %d", port);
+	syslog(LOG_NOTICE, "HTTP server listening on %s:%d", address, port);
 	syslog(LOG_NOTICE, "HTTP server root at %s", root_dir);
 	
 	/* Accept connections */
