@@ -2,32 +2,7 @@
  * Copyright (c) 2002, 2003 Niels Provos <provos@citi.umich.edu>
  * All rights reserved.
  * 
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * <LICENSEHERE>
  */
 
 #include <sys/types.h>
@@ -54,6 +29,7 @@
 
 #include <event.h>
 
+#include "network.h"
 #include "router.h"
 #include "pool.h"
 #include "interface.h"
@@ -92,7 +68,8 @@ tunnelcompare(struct router_entry *a, struct router_entry *b)
 
 	if ((res = addr_cmp(&a->tunnel_src, &b->tunnel_src)) != 0)
 		return (res);
-	return (addr_cmp(&a->tunnel_dst, &b->tunnel_dst));
+	else
+		return (addr_cmp(&a->tunnel_dst, &b->tunnel_dst));
 }
 
 SPLAY_PROTOTYPE(tunneltree, router_entry, node, tunnelcompare);
@@ -112,40 +89,6 @@ static struct pool *pool_network;
 static void network_addnode(struct network **, struct network *,
     struct network **);
 
-/*
- *  compare 2 v4 ranges
- */
-
-enum net_order
-network_compare(struct network *a, struct network *b)
-{
-	struct addr addr_a, addr_b;
-	struct addr addr_aend, addr_bend;
-
-	/* Set up the addresses; still IPv4 dependent */
-	addr_a = a->net;
-	addr_a.addr_bits = IP_ADDR_BITS;
-	addr_b = b->net;
-	addr_b.addr_bits = IP_ADDR_BITS;
-
-	addr_bcast(&a->net, &addr_aend);
-	addr_aend.addr_bits = IP_ADDR_BITS;
-	addr_bcast(&b->net, &addr_bend);
-	addr_bend.addr_bits = IP_ADDR_BITS;
-
-	if (addr_cmp(&addr_aend, &addr_b) < 0)
-		return (NET_PRECEEDS);
-	if (addr_cmp(&addr_a, &addr_bend) > 0)
-		return (NET_FOLLOWS);
-	if (addr_cmp(&addr_a, &addr_b) <= 0 && 
-	    addr_cmp(&addr_aend, &addr_bend) >= 0){
-		if (addr_cmp(&a->net, &b->net) == 0)
-			return (NET_EQUALS);
-		return (NET_CONTAINS);
-	}
-	return (NET_CONTAINED);
-}
-
 static void
 network_copydata(struct network *dst, struct network *src)
 {
