@@ -2,32 +2,7 @@
  * Copyright (c) 2003, 2004 Niels Provos <provos@citi.umich.edu>
  * All rights reserved.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * <LICENSEHERE>
  */
 
 #include <sys/types.h>
@@ -75,20 +50,41 @@ static void stats_make_fd(struct addr *, u_short);
 static void stats_activate(struct stats *stats);
 static void stats_deactivate(struct stats *stats);
 
-static char *user_name;
-static char *user_key;
-static struct addr *user_dst;
-static u_short user_port;
-static int stats_fd;
-static struct event ev_send;
+/* Many static variables.  We don't like them */
 
-static struct measurement measurement;
-static struct timeval tv_start;
-static struct event ev_measure;
-static struct evbuffer *evbuf_measure;
-static struct evbuffer *evbuf_tmp;
+/* We might have other consumers of our created records */
+struct statscb {
+	TAILQ_ENTRY(statscb) next;
+
+	int (*cb)(const struct record *, void *);
+	void *cb_arg;
+};
+
+struct statscontrol {
+	char *user_name;
+	char *user_key;
+	struct addr *user_dst;
+	u_short user_port;
+	int stats_fd;
+	struct event ev_send;
+
+	struct measurement measurement;
+	struct timeval tv_start;
+	struct event ev_measure;
+	struct evbuffer *evbuf_measure;
+	struct evbuffer *evbuf_tmp;
+
+	struct hmac_state hmac;
+
+	TAILQ_HEAD(statscbq, statscb) callbacks;
+
+	TAILQ_HEAD(statspackets, stats_packet) send_queue;
+	TAILQ_HEAD(statsqueue, stats) active_stats;
+	SPLAY_HEAD(statstree, stats) all_stats;
+};
+
+struct statscontrol sc;
 
-struct hmac_state hmac;
 
 static int
 compare(struct stats *a, struct stats *b)
@@ -96,10 +92,6 @@ compare(struct stats *a, struct stats *b)
 	return (conhdr_compare(&a->conhdr, &b->conhdr));
 }
 
-TAILQ_HEAD(statspackets, stats_packet) send_queue;
-TAILQ_HEAD(statsqueue, stats) active_stats;
-SPLAY_HEAD(statstree, stats) all_stats;
-
 SPLAY_PROTOTYPE(statstree, stats, node, compare);
 SPLAY_GENERATE(statstree, stats, node, compare);
 
@@ -294,7 +286,7 @@ stats_measure_timeout(void)
 	uint32_t diff_ms;
 
 	gettimeofday(&now, NULL);
-	timersub(&now, &tv_start, &now);
+	timersub(&now, &sc.tv_start, &now);
 	diff_ms = (now.tv_sec * 1000) + (now.tv_usec / 1000);
 	diff_ms %= (STATS_MEASUREMENT_INTERVAL * 1000);
 	diff_ms = (STATS_MEASUREMENT_INTERVAL * 1000) - diff_ms;
@@ -302,7 +294,7 @@ stats_measure_timeout(void)
 	tv.tv_sec = diff_ms / 1000;
 	tv.tv_usec = diff_ms * 1000;
 
-	evtimer_add(&ev_measure, &tv);
+	evtimer_add(&sc.ev_measure, &tv);
 }
 
 
@@ -321,7 +313,7 @@ stats_find(const struct tuple *conhdr)
 {
 	struct stats tmp, *res;
 	tmp.conhdr = *conhdr;
-	res = SPLAY_FIND(statstree, &all_stats, &tmp);
+	res = SPLAY_FIND(statstree, &sc.all_stats, &tmp);
 
 	if (res != NULL)
 		stats_add_timeout(res);
@@ -335,7 +327,7 @@ stats_activate(struct stats *stats)
 	if (stats->isactive)
 		return;
 	stats->isactive = 1;
-	TAILQ_INSERT_TAIL(&active_stats, stats, next);
+	TAILQ_INSERT_TAIL(&sc.active_stats, stats, next);
 }
 
 static void
@@ -344,7 +336,7 @@ stats_reactivate(struct stats *stats)
 	if (stats->isactive)
 		return;
 	stats->isactive = 1;
-	TAILQ_INSERT_HEAD(&active_stats, stats, next);
+	TAILQ_INSERT_HEAD(&sc.active_stats, stats, next);
 }
 
 static void
@@ -357,7 +349,7 @@ stats_deactivate(struct stats *stats)
 	stats->record.state &= ~RECORD_STATE_NEW;
 
 	stats->isactive = 0;
-	TAILQ_REMOVE(&active_stats, stats, next);
+	TAILQ_REMOVE(&sc.active_stats, stats, next);
 }
 
 static void
@@ -365,8 +357,8 @@ stats_ready_cb(int fd, short what, void *arg)
 {
 	struct stats_packet *tmp;
 
-	tmp = TAILQ_FIRST(&send_queue);
-	TAILQ_REMOVE(&send_queue, tmp, next);
+	tmp = TAILQ_FIRST(&sc.send_queue);
+	TAILQ_REMOVE(&sc.send_queue, tmp, next);
 	if (what != EV_TIMEOUT) {
 		syslog(LOG_DEBUG, "writing stats of length %d",
 		    EVBUFFER_LENGTH(tmp->evbuf));
@@ -374,7 +366,7 @@ stats_ready_cb(int fd, short what, void *arg)
 			syslog(LOG_WARNING,
 			    "remote stats daemon unreachable: %m");
 			close(fd);
-			stats_make_fd(user_dst, user_port);
+			stats_make_fd(sc.user_dst, sc.user_port);
 		}
 	}
 
@@ -382,11 +374,11 @@ stats_ready_cb(int fd, short what, void *arg)
 	evbuffer_free(tmp->evbuf);
 	free(tmp);
 
-	if (TAILQ_FIRST(&send_queue) != NULL) {
+	if (TAILQ_FIRST(&sc.send_queue) != NULL) {
 		struct timeval tv;
 		timerclear(&tv);
 		tv.tv_sec = STATS_SEND_TIMEOUT;
-		event_add(&ev_send, &tv);
+		event_add(&sc.ev_send, &tv);
 	}
 }
 
@@ -395,16 +387,18 @@ stats_prepare_send(struct evbuffer *evbuf)
 {
 	struct timeval tv;
 	struct stats_packet *tmp;
+
+	assert(sc.stats_fd != -1);
 	
 	if ((tmp = calloc(1, sizeof(struct stats_packet))) == NULL)
 		err(1, "%s: calloc", __func__);
 
 	tmp->evbuf = evbuf;
-	TAILQ_INSERT_TAIL(&send_queue, tmp, next);
+	TAILQ_INSERT_TAIL(&sc.send_queue, tmp, next);
 
 	timerclear(&tv);
 	tv.tv_sec = STATS_SEND_TIMEOUT;
-	event_add(&ev_send, &tv);
+	event_add(&sc.ev_send, &tv);
 }
 
 static void
@@ -413,21 +407,27 @@ stats_package_measurement()
 	struct evbuffer *evbuf;
 	u_char digest[SHA1_DIGESTSIZE];
 
+	/* Do not send any file data when we don't have a collector defined */
+	if (sc.stats_fd == -1)
+		return;
+	
 	if ((evbuf = evbuffer_new()) == NULL)
 		err(1, "%s: evbuffer_new", __func__);
 
 	/* Compress the measured data */
-	stats_compress(evbuf_measure);
+	stats_compress(sc.evbuf_measure);
 
 	/* Sign the data - at this point, we could use compression */
-	hmac_sign(&hmac, digest, sizeof(digest),
-	    EVBUFFER_DATA(evbuf_measure), EVBUFFER_LENGTH(evbuf_measure));
+	hmac_sign(&sc.hmac, digest, sizeof(digest),
+	    EVBUFFER_DATA(sc.evbuf_measure),
+	    EVBUFFER_LENGTH(sc.evbuf_measure));
 
 	/* Create the signed buffer */
-	tag_marshal_string(evbuf, SIG_NAME, user_name);
+	tag_marshal_string(evbuf, SIG_NAME, sc.user_name);
 	tag_marshal(evbuf, SIG_DIGEST, digest, sizeof(digest));
 	tag_marshal(evbuf, SIG_COMPRESSED_DATA,
-	    EVBUFFER_DATA(evbuf_measure), EVBUFFER_LENGTH(evbuf_measure));
+	    EVBUFFER_DATA(sc.evbuf_measure),
+	    EVBUFFER_LENGTH(sc.evbuf_measure));
 
 	stats_prepare_send(evbuf);
 }
@@ -435,9 +435,9 @@ stats_package_measurement()
 void
 measurement_marshal(struct evbuffer *evbuf, struct measurement *m)
 {
-	tag_marshal_int(evbuf_measure, M_COUNTER, m->counter);
-	tag_marshal_timeval(evbuf_measure, M_TV_START, &m->tv_start);
-	tag_marshal_timeval(evbuf_measure, M_TV_END, &m->tv_end);
+	tag_marshal_int(sc.evbuf_measure, M_COUNTER, m->counter);
+	tag_marshal_timeval(sc.evbuf_measure, M_TV_START, &m->tv_start);
+	tag_marshal_timeval(sc.evbuf_measure, M_TV_END, &m->tv_end);
 }
 
 /*
@@ -448,18 +448,19 @@ static void
 stats_measure_cb(int fd, short what, void *arg)
 {
 	struct stats *stats;
-
+	struct statscb *statscb;
+	
 	/* Schedule a new timeout */
 	stats_measure_timeout();
 
-	gettimeofday(&measurement.tv_end, NULL);
+	gettimeofday(&sc.measurement.tv_end, NULL);
 
-	while ((stats = TAILQ_FIRST(&active_stats)) != NULL) {
-		evbuffer_drain(evbuf_measure, -1);
-		measurement.counter++;
-		measurement_marshal(evbuf_measure, &measurement);
-		while ((stats = TAILQ_FIRST(&active_stats)) != NULL &&
-		    EVBUFFER_LENGTH(evbuf_measure) < STATS_MAX_SIZE) {
+	while ((stats = TAILQ_FIRST(&sc.active_stats)) != NULL) {
+		evbuffer_drain(sc.evbuf_measure, -1);
+		sc.measurement.counter++;
+		measurement_marshal(sc.evbuf_measure, &sc.measurement);
+		while ((stats = TAILQ_FIRST(&sc.active_stats)) != NULL &&
+		    EVBUFFER_LENGTH(sc.evbuf_measure) < STATS_MAX_SIZE) {
 			struct hash *hash;
 			int i;
 
@@ -490,13 +491,24 @@ stats_measure_cb(int fd, short what, void *arg)
 				    next);
 			}
 
+			/*
+			 * Check if we have any external consumers of
+			 * our records.
+			 */
+			TAILQ_FOREACH(statscb, &sc.callbacks, next) {
+				if ((*statscb->cb)(&stats->record,
+					statscb->cb_arg) == 1)
+					break;
+			}
+			
 			/*
 			 * Add to temporary buffer first, so that we can 
 			 * check our size contraints.
 			 */
 
-			evbuffer_drain(evbuf_tmp, -1);
-			tag_marshal_record(evbuf_tmp, M_RECORD,&stats->record);
+			evbuffer_drain(sc.evbuf_tmp, -1);
+			tag_marshal_record(sc.evbuf_tmp, M_RECORD,
+			    &stats->record);
 
 			/* Remove data that we have reported */
 			record_remove_hashes(&stats->record.hashes);
@@ -520,8 +532,8 @@ stats_measure_cb(int fd, short what, void *arg)
 			if (stats->needelete && !stats->isactive)
 				stats_free(stats);
 
-			if (EVBUFFER_LENGTH(evbuf_measure) +
-			    EVBUFFER_LENGTH(evbuf_tmp) >= STATS_MAX_SIZE) {
+			if (EVBUFFER_LENGTH(sc.evbuf_measure) +
+			    EVBUFFER_LENGTH(sc.evbuf_tmp) >= STATS_MAX_SIZE) {
 				/* Package up current packet */
 				stats_package_measurement();
 
@@ -529,20 +541,20 @@ stats_measure_cb(int fd, short what, void *arg)
 				 * Now clear the buffer and prepare it for
 				 * more stats.
 				 */
-				evbuffer_drain(evbuf_measure, -1);
-				measurement_marshal(evbuf_measure,
-				    &measurement);
+				evbuffer_drain(sc.evbuf_measure, -1);
+				measurement_marshal(sc.evbuf_measure,
+				    &sc.measurement);
 			}
 
-			evbuffer_add_buffer(evbuf_measure, evbuf_tmp);
+			evbuffer_add_buffer(sc.evbuf_measure, sc.evbuf_tmp);
 		}
 
 		stats_package_measurement();
 	}
 
 	/* Start the next measuring period */
-	measurement.tv_start = measurement.tv_end;
-	timerclear(&measurement.tv_end);
+	sc.measurement.tv_start = sc.measurement.tv_end;
+	timerclear(&sc.measurement.tv_end);
 }
 
 static void
@@ -578,7 +590,7 @@ stats_new(const struct tuple *conhdr)
 
 	stats->record.state = RECORD_STATE_NEW;
 
-	SPLAY_INSERT(statstree, &all_stats, stats);
+	SPLAY_INSERT(statstree, &sc.all_stats, stats);
 	stats_activate(stats);
 
 	return (stats);
@@ -636,6 +648,13 @@ record_fill(struct record *r, const struct tuple *hdr)
         name = honeyd_osfp_name(&ip);
 	if (name != NULL)
 		r->os_fp = strdup(name);
+
+	/*
+	 * Mark this connection as one that originated locally.  This will
+	 * allow our stats measurement mechanisms to ignore it.
+	 */
+	if (hdr->local)
+		r->flags |= REC_FLAG_LOCAL;
 }
 
 void
@@ -663,7 +682,7 @@ record_clean(struct record *record)
 void
 stats_free(struct stats *stats)
 {
-	SPLAY_REMOVE(statstree, &all_stats, stats);
+	SPLAY_REMOVE(statstree, &sc.all_stats, stats);
 	stats_deactivate(stats);
 
 	evtimer_del(&stats->ev_timeout);
@@ -674,6 +693,24 @@ stats_free(struct stats *stats)
 	free(stats);
 }
 
+/* Functions for processing incoming network data */
+
+static void
+stats_process_data(struct stats *stats, void *data, u_int len)
+{
+	if (data == NULL) {
+		/* This object has been terminated */
+		gettimeofday(&stats->record.tv_end, NULL);
+		stats->needelete = 1;
+		return;
+	}
+
+	stats->record.bytes += len;
+
+	evbuffer_add(stats->evbuf, data, len);
+	stats_shingle_data(stats);
+}
+
 static void
 stats_tcp_input(struct tuple *conhdr, u_char *pkt, u_int pktlen, void *arg)
 {
@@ -702,22 +739,6 @@ stats_udp_input(struct tuple *conhdr, u_char *pkt, u_int pktlen, void *arg)
 		stats = stats_new(conhdr);
 }
 
-static void
-stats_process_data(struct stats *stats, void *data, u_int len)
-{
-	if (data == NULL) {
-		/* This object has been terminated */
-		gettimeofday(&stats->record.tv_end, NULL);
-		stats->needelete = 1;
-		return;
-	}
-
-	stats->record.bytes += len;
-
-	evbuffer_add(stats->evbuf, data, len);
-	stats_shingle_data(stats);
-}
-
 static void
 stats_tcp_data(struct tuple *conhdr, u_char *pkt, u_int pktlen, void *arg)
 {
@@ -741,25 +762,44 @@ stats_udp_data(struct tuple *conhdr, u_char *pkt, u_int pktlen, void *arg)
 static void
 stats_make_fd(struct addr *dst, u_short port)
 {
-	stats_fd = make_socket(connect, SOCK_DGRAM, addr_ntoa(dst), port);
-	if (stats_fd == -1)
+	sc.stats_fd = make_socket(connect, SOCK_DGRAM, addr_ntoa(dst), port);
+	if (sc.stats_fd == -1)
 		err(1, "%s: make_socket", __func__);
-	event_set(&ev_send, stats_fd, EV_WRITE, stats_ready_cb, NULL);
+	event_set(&sc.ev_send, sc.stats_fd, EV_WRITE, stats_ready_cb, NULL);
 }
 
 void
-stats_init(struct addr *dst, u_short port, char *name, char *password)
+stats_register_cb(int (*cb)(const struct record *, void *), void *cb_arg)
 {
-	/* Information to establish the authentication */
-	user_name = name;
-	user_key = password;
-	user_dst = dst;
-	user_port = port;
+	struct statscb *statscb = calloc(1, sizeof(struct statscb));
+	assert(statscb != NULL);
+
+	statscb->cb = cb;
+	statscb->cb_arg = cb_arg;
+
+	TAILQ_INSERT_TAIL(&sc.callbacks, statscb, next);
+}
+
+void
+stats_init_collect(struct addr *dst, u_short port, char *name, char *password)
+{
+	sc.user_name = name;
+	sc.user_key = password;
+	sc.user_dst = dst;
+	sc.user_port = port;
 
 	stats_make_fd(dst, port);
 
 	/* Set up message authentication code */
-	hmac_init(&hmac, user_key);
+	hmac_init(&sc.hmac, sc.user_key);
+}
+
+void
+stats_init()
+{
+	/* Information to establish the authentication */
+	memset(&sc, 0, sizeof(sc));
+	sc.stats_fd = -1;
 
 	/* Setup hooks that we use for data processing */
 	hooks_add_packet_hook(IP_PROTO_TCP, HD_INCOMING,
@@ -772,19 +812,21 @@ stats_init(struct addr *dst, u_short port, char *name, char *password)
 	hooks_add_packet_hook(IP_PROTO_UDP, HD_INCOMING_STREAM,
 	    stats_udp_data, NULL);
 
-	TAILQ_INIT(&send_queue);
+	TAILQ_INIT(&sc.callbacks);
+	
+	TAILQ_INIT(&sc.send_queue);
 
-	TAILQ_INIT(&active_stats);
-	SPLAY_INIT(&all_stats);
+	TAILQ_INIT(&sc.active_stats);
+	SPLAY_INIT(&sc.all_stats);
 
-	evbuf_measure = evbuffer_new();
-	evbuf_tmp = evbuffer_new();
+	sc.evbuf_measure = evbuffer_new();
+	sc.evbuf_tmp = evbuffer_new();
 
 	/* Let the measurements begin */
-	memset(&measurement, 0, sizeof(measurement));
-	gettimeofday(&measurement.tv_start, NULL);
-	tv_start = measurement.tv_start;
-	evtimer_set(&ev_measure, stats_measure_cb, NULL);
+	memset(&sc.measurement, 0, sizeof(sc.measurement));
+	gettimeofday(&sc.measurement.tv_start, NULL);
+	sc.tv_start = sc.measurement.tv_start;
+	evtimer_set(&sc.ev_measure, stats_measure_cb, NULL);
 
 	stats_measure_timeout();
 }
@@ -795,13 +837,15 @@ stats_hmac_test()
 	u_char digest[SHA1_DIGESTSIZE];
 	char *test1 = "test", *test2 = "txst";
 
-	hmac_init(&hmac, "1234");
-	hmac_sign(&hmac, digest, sizeof(digest), test1, strlen(test1));
+	hmac_init(&sc.hmac, "1234");
+	hmac_sign(&sc.hmac, digest, sizeof(digest), test1, strlen(test1));
 
-	if (!hmac_verify(&hmac, digest, sizeof(digest), test1, strlen(test1)))
+	if (!hmac_verify(&sc.hmac, digest, sizeof(digest),
+		test1, strlen(test1)))
 		errx(1, "%s: verify failed", __func__);
 
-	if (hmac_verify(&hmac, digest, sizeof(digest), test2, strlen(test2)))
+	if (hmac_verify(&sc.hmac, digest, sizeof(digest),
+		test2, strlen(test2)))
 		errx(1, "%s: verify should have failed", __func__);
 
 	fprintf(stderr, "\t%s: OK\n", __func__);
