@@ -2,32 +2,7 @@
  * Copyright (c) 2001, 2004 Niels Provos <provos@citi.umich.edu>
  * All rights reserved.
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * <LICENSEHERE>
  */
 
 #include <sys/types.h>
@@ -55,9 +30,13 @@
 #include "config.h"
 
 #include <event.h>
+#include <dnet.h>
 
 #include "ui.h"
 #include "parser.h"
+#ifdef HAVE_PYTHON
+#include "pyextend.h"
+#endif
 
 char *ui_file = UI_FIFO;
 
@@ -67,6 +46,7 @@ char *ui_file = UI_FIFO;
 char *strnsep(char **, char *);
 
 int ui_command_help(struct evbuffer *, char *);
+int ui_command_python(struct evbuffer *, char *);
 
 struct command {
 	char *cmd;
@@ -82,11 +62,22 @@ struct command commands[] = {
 		"help [command]\n",
 		ui_command_help
 	},
+	{
+		"!",
+		"!\t\t runs a Python command in the Honeyd environment\n",
+		"! <command >",
+		ui_command_python
+	},
 	{
 		"delete",
 		"delete\t\t removes configured templates and ports\n",
 		"delete <template|template proto port number>\n",
 	},
+	{
+		"list",
+		"list\t\t lists configured templates or subsystems\n",
+		"list <template [pattern]|subsystem [pattern]>\n",
+	},
 	{
 		NULL, NULL, NULL, NULL
 	}
@@ -144,6 +135,19 @@ ui_dead(struct uiclient *client)
 	free(client);
 }
 
+int
+ui_command_python(struct evbuffer *buf, char *line)
+{
+#ifndef HAVE_PYTHON
+	const char *error_python = 
+	    "Error: Honeyd has been compiled without Python support.\n";
+	evbuffer_add(buf, error_python, strlen(error_python));
+#else
+	pyextend_run(buf, line);
+#endif
+	return (0);
+}
+
 int
 ui_command_help(struct evbuffer *buf, char *line)
 {
